import os
import unittest
from unittest import TestCase
import lzma
import pickle
import numpy as np
import scipy.sparse

from clayrs.content_analyzer.exogenous_properties_retrieval import PropertiesFromDataset
from clayrs.content_analyzer import ContentAnalyzer, FieldConfig, ExogenousConfig, ItemAnalyzerConfig
from clayrs.content_analyzer.content_representation.content import SimpleField, FeaturesBagField, \
    EmbeddingField, IndexField, PropertiesDict
from clayrs.content_analyzer.field_content_production_techniques import OriginalData
from clayrs.content_analyzer.embeddings.embedding_loader.gensim import Gensim
from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.embedding_technique \
    import WordEmbeddingTechnique
from clayrs.content_analyzer.field_content_production_techniques.tf_idf import SkLearnTfIdf
from clayrs.content_analyzer.information_processor import NLTK
from clayrs.content_analyzer.memory_interfaces import SearchIndex, KeywordIndex
from clayrs.content_analyzer.raw_information_source import JSONFile
from test import dir_test_files

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
movies_info_reduced = os.path.join(dir_test_files, "movies_info_reduced.json")

decode_path = os.path.join(dir_test_files, "test_decode/")
decode_string = os.path.join(decode_path, "movies_title_string.json")
decode_tfidf = os.path.join(decode_path, "movies_title_tfidf.json")
decode_embedding = os.path.join(decode_path, "movies_title_embedding.json")


class TestContentsProducer(TestCase):
    def test_create_content(self):
        exogenous_config = ExogenousConfig(PropertiesFromDataset(field_name_list=['Title']))
        content_analyzer_config = ItemAnalyzerConfig(JSONFile(movies_info_reduced), ["imdbID"], "movielens_test")
        content_analyzer_config.add_single_exogenous(exogenous_config)
        content_analyzer_config.add_single_exogenous(ExogenousConfig(PropertiesFromDataset(field_name_list=['Title'])))
        content_analyzer = ContentAnalyzer(content_analyzer_config)
        content_analyzer.fit()

        for name in os.listdir(THIS_DIR):
            if os.path.isdir(os.path.join(THIS_DIR, name)) \
                    and 'movielens_test' in str(name):

                with lzma.open(os.path.join(THIS_DIR, name, 'tt0113497.xz'), 'r') as file:
                    content = pickle.load(file)

                    self.assertIsInstance(content.get_exogenous_representation(0), PropertiesDict)
                    self.assertIsInstance(content.get_exogenous_representation(0).value, dict)
                    break

    def test_field_exceptions(self):
        # test to make sure that the method that checks the field configs ids for each field name in the field_dict
        # of the content analyzer works. It considers the three cases this can occur: when passing the field_dict
        # with duplicate ids as argument for the content_analyzer, when setting the FieldConfig list with duplicates
        # for a specific field_name, and when appending a FieldConfig to the list associated with a specific field_name
        # but the config id is already in the list

        config_1 = FieldConfig(SkLearnTfIdf(), NLTK(), id="test")
        config_2 = FieldConfig(SkLearnTfIdf(), NLTK(), id="test")
        config_list = [config_1, config_2]
        field_dict = dict()
        field_dict["test"] = config_list

        with self.assertRaises(ValueError):
            config = ItemAnalyzerConfig(JSONFile(movies_info_reduced), ["imdbID"], "movielens_test", field_dict)
            ContentAnalyzer(config).fit()

        with self.assertRaises(ValueError):
            config = ItemAnalyzerConfig(JSONFile(movies_info_reduced), ["imdbID"], "movielens_test")
            config.add_multiple_config("test", config_list)
            ContentAnalyzer(config).fit()

        with self.assertRaises(ValueError):
            config = ItemAnalyzerConfig(JSONFile(movies_info_reduced), ["imdbID"], "movielens_test")
            config.add_single_config("test", config_1)
            config.add_single_config("test", config_2)
            ContentAnalyzer(config).fit()

    def test_exogenous_exceptions(self):
        # test to make sure that the method that checks the exogenous configs ids in the exogenous_representation_list
        # of the content analyzer works. It considers the two cases this can occur: when passing the
        # exogenous_representation_list with duplicate ids as argument for the content_analyzer,
        # and when appending an ExogenousConfig to the list but the config id is already in the list

        config_1 = ExogenousConfig(PropertiesFromDataset(field_name_list=['Title']), "test")
        config_2 = ExogenousConfig(PropertiesFromDataset(field_name_list=['Title']), "test")
        exogenous_representation_list = [config_1, config_2]

        with self.assertRaises(ValueError):
            config = ItemAnalyzerConfig(JSONFile(movies_info_reduced), ["imdbID"], "movielens_test",
                                        exogenous_representation_list=exogenous_representation_list)
            ContentAnalyzer(config).fit()

        with self.assertRaises(ValueError):
            config = ItemAnalyzerConfig(JSONFile(movies_info_reduced), ["imdbID"], "movielens_test")
            config.add_single_exogenous(config_1)
            config.add_single_exogenous(config_2)
            ContentAnalyzer(config).fit()

    def test_create_content_tfidf(self):
        movies_ca_config = ItemAnalyzerConfig(
            source=JSONFile(movies_info_reduced),
            id='imdbID',
            output_directory="movielens_test_tfidf",
        )

        movies_ca_config.add_multiple_config(
            field_name='Title',
            config_list=[FieldConfig(SkLearnTfIdf())])

        content_analyzer = ContentAnalyzer(movies_ca_config)
        content_analyzer.fit()

        for name in os.listdir(THIS_DIR):
            if os.path.isdir(os.path.join(THIS_DIR, name)) \
                    and 'movielens_test_tfidf' in str(name):

                with lzma.open(os.path.join(THIS_DIR, name, 'tt0113497.xz'), 'r') as file:
                    content = pickle.load(file)

                    self.assertIsInstance(content.get_field("Title")[0], FeaturesBagField)
                    self.assertIsInstance(content.get_field("Title")[0].value, scipy.sparse.csc_matrix)
                    break

    def test_create_content_embedding(self):
        movies_ca_config = ItemAnalyzerConfig(
            source=JSONFile(movies_info_reduced),
            id=['imdbID'],
            output_directory="movielens_test_embedding",
        )

        movies_ca_config.add_multiple_config(
            field_name='Title',
            config_list=[FieldConfig(
                    WordEmbeddingTechnique(Gensim('glove-wiki-gigaword-50')),
                    NLTK(lemmatization=True, stopwords_removal=True))])

        content_analyzer = ContentAnalyzer(movies_ca_config)
        content_analyzer.fit()

        for name in os.listdir(THIS_DIR):
            if os.path.isdir(os.path.join(THIS_DIR, name)) \
                    and 'movielens_test_embedding' in str(name):

                with lzma.open(os.path.join(THIS_DIR, name, 'tt0113497.xz'), 'r') as file:
                    content = pickle.load(file)

                    self.assertIsInstance(content.get_field("Title")[0], EmbeddingField)
                    self.assertIsInstance(content.get_field("Title")[0].value, np.ndarray)
                    break

    def test_create_contents_in_index(self):
        output_dir = os.path.join(THIS_DIR, "movielens_test_original_index")
        movies_ca_config = ItemAnalyzerConfig(
            source=JSONFile(movies_info_reduced),
            id=['imdbID'],
            output_directory=output_dir,
        )

        movies_ca_config.add_multiple_config(
            field_name='Title',
            config_list=[FieldConfig(OriginalData(), NLTK(lemmatization=True, stopwords_removal=True),
                         SearchIndex(os.path.join(output_dir, "index")), "test_search"),

                         FieldConfig(SkLearnTfIdf(), NLTK(),
                                     KeywordIndex(os.path.join(output_dir, "index1")),
                                     "test_keyword"),

                         FieldConfig(OriginalData(), NLTK(),
                                     SearchIndex(os.path.join(output_dir, "index")))
                         ])

        content_analyzer = ContentAnalyzer(movies_ca_config)
        content_analyzer.fit()

        for name in os.listdir(THIS_DIR):
            if os.path.isdir(os.path.join(THIS_DIR, name)) \
                    and 'movielens_test_original_index' in str(name):

                with lzma.open(os.path.join(THIS_DIR, name, 'tt0113497.xz'), 'r') as file:
                    content = pickle.load(file)

                    self.assertIsInstance(content.get_field("Title")[0], IndexField)
                    self.assertIsInstance(content.get_field("Title")[0].value, str)
                    self.assertIsInstance(content.get_field("Title")[1], IndexField)
                    self.assertIsInstance(content.get_field("Title")[1].value, str)
                    break

    # Functionality to decode NOT IMPLEMENTED
    #
    # def test_decode_field_data_string(self):
    #     movies_ca_config = ItemAnalyzerConfig(
    #         source=JSONFile(decode_string),
    #         id=['imdbID'],
    #         output_directory=decode_path + 'movies_string_'
    #     )
    #
    #     movies_ca_config.add_multiple_config(
    #         field_name='Title',
    #         config_list=[FieldConfig()]
    #     )
    #     ContentAnalyzer(config=movies_ca_config).fit()
    #
    #     for name in os.listdir(decode_path):
    #         if os.path.isdir(os.path.join(decode_path, name)) \
    #                 and 'movies_string_' in str(name):
    #
    #             with lzma.open(os.path.join(decode_path, name, 'tt0113497.xz'), 'r') as file:
    #                 content = pickle.load(file)
    #
    #                 self.assertIsInstance(content.get_field("Title")[0], SimpleField)
    #                 self.assertIsInstance(content.get_field("Title")[0].value, str)
    #                 break
    #
    # def test_decode_field_data_tfidf(self):
    #     movies_ca_config = ItemAnalyzerConfig(
    #         source=JSONFile(decode_tfidf),
    #         id=['imdbID'],
    #         output_directory=decode_path + 'movies_tfidf_'
    #     )
    #
    #     movies_ca_config.add_multiple_config(
    #         field_name='Title',
    #         config_list=[FieldConfig()]
    #     )
    #     ContentAnalyzer(config=movies_ca_config).fit()
    #
    #     for name in os.listdir(decode_path):
    #         if os.path.isdir(os.path.join(decode_path, name)) \
    #                 and 'movies_tfidf_' in str(name):
    #             with lzma.open(os.path.join(decode_path, name, 'tt0113497.xz'), 'r') as file:
    #                 content = pickle.load(file)
    #
    #                 self.assertIsInstance(content.get_field("Title")[0], FeaturesBagField)
    #                 self.assertIsInstance(content.get_field("Title")[0].value, dict)
    #                 break
    #
    # def test_decode_field_data_embedding(self):
    #     movies_ca_config = ItemAnalyzerConfig(
    #         source=JSONFile(decode_embedding),
    #         id=['imdbID'],
    #         output_directory=decode_path + 'movies_embedding_'
    #     )
    #
    #     movies_ca_config.add_multiple_config(
    #         field_name='Title',
    #         config_list=[FieldConfig()]
    #     )
    #     ContentAnalyzer(config=movies_ca_config).fit()
    #
    #     for name in os.listdir(decode_path):
    #         if os.path.isdir(os.path.join(decode_path, name)) \
    #                 and 'movies_embedding_' in str(name):
    #             with lzma.open(os.path.join(decode_path, name, 'tt0113497.xz'), 'r') as file:
    #                 content = pickle.load(file)
    #
    #                 self.assertIsInstance(content.get_field("Title")[0], EmbeddingField)
    #                 self.assertIsInstance(content.get_field("Title")[0].value, np.ndarray)
    #                 break


class TestContentAnalyzer(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.out_dir = 'test_export_json/'

    def test_fit_export_json(self):
        movies_ca_config = ItemAnalyzerConfig(
            source=JSONFile(movies_info_reduced),
            id=['imdbID'],
            output_directory=self.out_dir,
            export_json=True
        )

        movies_ca_config.add_single_config('Plot', FieldConfig(OriginalData()))
        movies_ca_config.add_single_config('Plot', FieldConfig(SkLearnTfIdf()))
        movies_ca_config.add_single_config('imdbRating', FieldConfig())

        ContentAnalyzer(movies_ca_config).fit()

        self.assertTrue(os.path.isfile(os.path.join(self.out_dir, 'contents.json')))
        processed_source = list(JSONFile(os.path.join(self.out_dir, 'contents.json')))

        self.assertEqual(len(processed_source), 20)
        for processed_content in processed_source:
            self.assertIn('Plot#0', processed_content)
            self.assertIn('Plot#1', processed_content)
            self.assertIn('imdbRating#0', processed_content)

    # def doCleanups(self) -> None:
    #     if os.path.isdir(self.out_dir):
    #         shutil.rmtree(self.out_dir)


if __name__ == "__main__":
    unittest.main()



import unittest
import pandas as pd
import numpy as np
from scipy.stats import ttest_ind, ranksums

from clayrs.evaluation.statistical_test import StatisticalTest, Ttest, Wilcoxon, PairedTest


class TestStatisticalTest(unittest.TestCase):
    def test__common_users(self):
        # all_users are in common
        equal_users_sys1 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                         'NDCG': [0.5, 0.7, 0.8, 0.2],
                                         'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        equal_users_sys2 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                         'NDCG': [0.8, 0.2, 0.2, 0.4],
                                         'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result = StatisticalTest._common_users(equal_users_sys1, equal_users_sys2, column_list=['NDCG'])

        expected_dict = {'NDCG_x': [0.5, 0.7, 0.8, 0.2],
                         'user_id': ['u1', 'u2', 'u3', 'u4'],
                         'NDCG_y': [0.8, 0.2, 0.2, 0.4]}
        result_dict = result.to_dict(orient='list')

        self.assertDictEqual(expected_dict, result_dict)

        # not all user are in common
        equal_users_sys1 = pd.DataFrame({'user_id': ['u5', 'u6', 'u3', 'u4'],
                                         'NDCG': [0.5, 0.7, 0.8, 0.2],
                                         'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        equal_users_sys2 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                         'NDCG': [0.8, 0.2, 0.2, 0.4],
                                         'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result = StatisticalTest._common_users(equal_users_sys1, equal_users_sys2, column_list=['NDCG'])

        expected_dict = {'NDCG_x': [0.8, 0.2],
                         'user_id': ['u3', 'u4'],
                         'NDCG_y': [0.2, 0.4]}
        result_dict = result.to_dict(orient='list')

        self.assertDictEqual(expected_dict, result_dict)

        # no metric in common
        equal_users_sys1 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                         'MRR': [0.5, 0.7, 0.8, 0.2],
                                         'Precision - macro': [0.998, 0.123, 0.556, 0.887]})

        equal_users_sys2 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                         'NDCG': [0.8, 0.2, 0.2, 0.4],
                                         'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result = StatisticalTest._common_users(equal_users_sys1, equal_users_sys2, column_list=[])

        expected_dict = {'user_id': ['u1', 'u2', 'u3', 'u4']}
        result_dict = result.to_dict(orient='list')

        self.assertDictEqual(expected_dict, result_dict)

        # no metric in common and no user in common
        equal_users_sys1 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                         'MRR': [0.5, 0.7, 0.8, 0.2],
                                         'Precision - macro': [0.998, 0.123, 0.556, 0.887]})

        equal_users_sys2 = pd.DataFrame({'user_id': ['u5', 'u6', 'u7', 'u8'],
                                         'NDCG': [0.8, 0.2, 0.2, 0.4],
                                         'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result = StatisticalTest._common_users(equal_users_sys1, equal_users_sys2, column_list=[])

        expected_dict = {'user_id': []}
        result_dict = result.to_dict(orient='list')

        self.assertDictEqual(expected_dict, result_dict)


class TestPairedTest(unittest.TestCase):
    """
    Contains all methods on which a paired test must be tested.
    The class_to_test parameter is basically the class of the framework deriving PairedTest that
    we want to test
    The 'external_function' parameter is the function used to implement the class, usually taken from an external
    library
    EXAMPLE:
            class_to_test = Ttest()
            external_function = scipy.stats.ttest_ind
    """

    def perform_all_in_common(self, class_to_test, external_function):
        # perform all user and all columns in common
        users_metrics_result1 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                              'NDCG': [0.5, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                              'NDCG': [0.8, 0.2, 0.2, 0.4],
                                              'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(['NDCG', 'Precision - micro'] == list(result_df.columns))

        result_ndcg = result_df['NDCG'].values[0]
        result_precision = result_df['Precision - micro'].values[0]

        expected_ndcg = external_function([0.5, 0.7, 0.8, 0.2], [0.8, 0.2, 0.2, 0.4])
        expected_precision = external_function([0.998, 0.123, 0.556, 0.887], [0.45, 0.23, 0.112, 0.776])

        self.assertEqual(expected_ndcg, result_ndcg)
        self.assertEqual(expected_precision, result_precision)

    def perform_3_systems(self, class_to_test, external_function):
        # perform all user and all columns in common
        users_metrics_result1 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                              'NDCG': [0.5, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                              'NDCG': [0.8, 0.2, 0.2, 0.4],
                                              'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        users_metrics_result3 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                              'NDCG': [0.9, 0.3, 0.2, 0.5],
                                              'Precision - micro': [0.44, 0.88, 0.21, 0.56]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2, users_metrics_result3])

        self.assertTrue(['NDCG', 'Precision - micro'] == list(result_df.columns))

        # we expect sys1/sys2, sys1/sys3, sys2/sys3
        self.assertTrue(len(result_df) == 3)

        result_ndcg_12 = result_df['NDCG'].values[0]
        result_ndcg_13 = result_df['NDCG'].values[1]
        result_ndcg_23 = result_df['NDCG'].values[2]
        result_precision_12 = result_df['Precision - micro'].values[0]
        result_precision_13 = result_df['Precision - micro'].values[1]
        result_precision_23 = result_df['Precision - micro'].values[2]

        expected_ndcg_12 = external_function([0.5, 0.7, 0.8, 0.2], [0.8, 0.2, 0.2, 0.4])
        expected_ndcg_13 = external_function([0.5, 0.7, 0.8, 0.2], [0.9, 0.3, 0.2, 0.5])
        expected_ndcg_23 = external_function([0.8, 0.2, 0.2, 0.4], [0.9, 0.3, 0.2, 0.5])
        expected_precision_12 = external_function([0.998, 0.123, 0.556, 0.887], [0.45, 0.23, 0.112, 0.776])
        expected_precision_13 = external_function([0.998, 0.123, 0.556, 0.887], [0.44, 0.88, 0.21, 0.56])
        expected_precision_23 = external_function([0.45, 0.23, 0.112, 0.776], [0.44, 0.88, 0.21, 0.56])

        self.assertEqual(expected_ndcg_12, result_ndcg_12)
        self.assertEqual(expected_ndcg_13, result_ndcg_13)
        self.assertEqual(expected_ndcg_23, result_ndcg_23)
        self.assertEqual(expected_precision_12, result_precision_12)
        self.assertEqual(expected_precision_13, result_precision_13)
        self.assertEqual(expected_precision_23, result_precision_23)

    def perform_only_some_in_common(self, class_to_test, external_function):
        # perform only some users in common
        users_metrics_result1 = pd.DataFrame({'user_id': ['u5', 'u6', 'u3', 'u4'],
                                              'NDCG': [0.5, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'NDCG': [0.8, 0.2, 0.2, 0.4],
                                              'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(['NDCG', 'Precision - micro'] == list(result_df.columns))

        result_ndcg = result_df['NDCG'].values[0]
        result_precision = result_df['Precision - micro'].values[0]

        expected_ndcg = external_function([0.8, 0.2], [0.2, 0.4])
        expected_precision = external_function([0.556, 0.887], [0.23, 0.776])

        self.assertEqual(expected_ndcg, result_ndcg)
        self.assertEqual(expected_precision, result_precision)

        # perform only some users and only some columns in common
        users_metrics_result1 = pd.DataFrame({'user_id': ['u5', 'u6', 'u3', 'u4'],
                                              'NDCG': [0.5, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'MRR': [0.8, 0.2, 0.2, 0.4],
                                              'NDCG': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(['NDCG'] == list(result_df.columns))

        result_ndcg = result_df['NDCG'].values[0]

        expected_ndcg = external_function([0.8, 0.2], [0.23, 0.776])

        self.assertEqual(expected_ndcg, result_ndcg)
        self.assertEqual(expected_precision, result_precision)

    def perform_nothing_in_common(self, class_to_test, external_function):
        # perform columns in common but no user in common
        users_metrics_result1 = pd.DataFrame({'user_id': ['u5', 'u6', 'u8', 'u9'],
                                              'NDCG': [0.5, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'NDCG': [0.8, 0.2, 0.2, 0.4],
                                              'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(['NDCG', 'Precision - micro'] == list(result_df.columns))

        result_ndcg = result_df['NDCG'].values[0]
        result_precision = result_df['Precision - micro'].values[0]

        self.assertTrue(np.isnan(result_ndcg.statistic))
        self.assertTrue(np.isnan(result_ndcg.pvalue))
        self.assertTrue(np.isnan(result_precision.statistic))
        self.assertTrue(np.isnan(result_precision.pvalue))

        # perform users in common but not columns in common
        users_metrics_result1 = pd.DataFrame({'user_id': ['u1', 'u2', 'u3', 'u4'],
                                              'NDCG': [0.5, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'MRR': [0.8, 0.2, 0.2, 0.4],
                                              'Recall': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(result_df.empty)

    def perform_nan_present(self, class_to_test, external_function):
        # u1 has nan values in NDCG column
        users_metrics_result1 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'NDCG': [np.nan, 0.7, 0.8, 0.2],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'NDCG': [np.nan, 0.2, 0.2, 0.4],
                                              'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(['NDCG', 'Precision - micro'] == list(result_df.columns))

        # ndcg column has one nan values, we don't use it
        expected_ndcg = external_function([0.7, 0.8, 0.2], [0.2, 0.2, 0.4])
        # other metric columns are unaffected
        expected_precision = external_function([0.998, 0.123, 0.556, 0.887], [0.45, 0.23, 0.112, 0.776])

        result_ndcg = result_df['NDCG'].values[0]
        result_precision = result_df['Precision - micro'].values[0]

        self.assertEqual(expected_ndcg, result_ndcg)
        self.assertEqual(expected_precision, result_precision)

        # all users have nan values in NDCG column
        users_metrics_result1 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'NDCG': [np.nan, np.nan, np.nan, np.nan],
                                              'Precision - micro': [0.998, 0.123, 0.556, 0.887]})

        users_metrics_result2 = pd.DataFrame({'user_id': ['u1', 'u3', 'u2', 'u4'],
                                              'NDCG': [np.nan, np.nan, np.nan, np.nan],
                                              'Precision - micro': [0.45, 0.23, 0.112, 0.776]})

        result_df = class_to_test.perform([users_metrics_result1, users_metrics_result2])

        self.assertTrue(['NDCG', 'Precision - micro'] == list(result_df.columns))

        # NDCG columns has all nan values, so we can't compute the statistic
        result_ndcg = result_df['NDCG'].values[0]

        self.assertTrue(np.isnan(result_ndcg.statistic))
        self.assertTrue(np.isnan(result_ndcg.pvalue))

        # other metric columns are unaffected
        expected_precision = external_function([0.998, 0.123, 0.556, 0.887], [0.45, 0.23, 0.112, 0.776])
        result_precision = result_df['Precision - micro'].values[0]

        self.assertEqual(expected_precision, result_precision)


class TestTtest(TestPairedTest):

    def test_perform_all_in_common(self):
        self.perform_all_in_common(Ttest(), ttest_ind)

    def test_perform_only_some_in_common(self):
        self.perform_only_some_in_common(Ttest(), ttest_ind)

    def test_perform_nothing_in_common(self):
        self.perform_nothing_in_common(Ttest(), ttest_ind)

    def test_perform_3_systems(self):
        self.perform_3_systems(Ttest(), ttest_ind)

    def test_perform_nan_present(self):
        self.perform_nan_present(Ttest(), ttest_ind)


class TestWilcoxon(TestPairedTest):

    def test_perform_all_in_common(self):
        self.perform_all_in_common(Wilcoxon(), ranksums)

    def test_perform_only_some_in_common(self):
        self.perform_only_some_in_common(Wilcoxon(), ranksums)

    def test_perform_nothing_in_common(self):
        self.perform_nothing_in_common(Wilcoxon(), ranksums)

    def test_perform_3_systems(self):
        self.perform_3_systems(Wilcoxon(), ranksums)

    def test_perform_nan_present(self):
        self.perform_nan_present(Wilcoxon(), ranksums)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict

import pandas as pd
import numpy as np
from typing import List, TYPE_CHECKING
from SPARQLWrapper import SPARQLWrapper, JSON, POST, GET
from SPARQLWrapper.SPARQLExceptions import URITooLong
from babelpy.babelfy import BabelfyClient

if TYPE_CHECKING:
    from clayrs.content_analyzer.raw_information_source import RawInformationSource
    from clayrs.content_analyzer.content_representation.content import ExogenousPropertiesRepresentation

from clayrs.content_analyzer.content_representation.content import PropertiesDict, EntitiesProp
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_progbar
from clayrs.content_analyzer.utils.check_tokenization import check_not_tokenized


class ExogenousPropertiesRetrieval(ABC):

    def __init__(self, mode: str = 'only_retrieved_evaluated'):
        """
        Class that creates a list of couples like this:
            <property: property value URI>
        The couples are properties retrieved from Linked Open Data Cloud

        Args:
            mode: one in: 'all', 'all_retrieved', 'only_retrieved_evaluated', 'original_retrieved',
        """
        self._check_mode(mode)
        self.__mode = mode

    @staticmethod
    def _check_mode(mode):
        modalities = {
            'all',
            'all_retrieved',
            'only_retrieved_evaluated',
            'original_retrieved',
        }
        if mode not in modalities:
            raise ValueError(f"mode={mode} not supported! Valid modalities are {modalities}")

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        self._check_mode(mode)
        self.__mode = mode

    @abstractmethod
    def get_properties(self, raw_source: RawInformationSource) -> List[ExogenousPropertiesRepresentation]:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class PropertiesFromDataset(ExogenousPropertiesRetrieval):
    """
    Exogenous technique which expands each content by using as external source the raw source itself

    Different modalities are available:

    * If `mode='only_retrieved_evaluated'` all fields for the content will be retrieved from raw source but discarding
    the ones with a blank value (i.e. '')

    ```title='JSON raw source'
    [{'Title': 'Jumanji', 'Year': 1995},
    {'Title': 'Toy Story', 'Year': ''}]
    ```

    ```python
    json_file = JSONFile(json_path)
    PropertiesFromDataset(mode='only_retrieved_evaluated').get_properties(json_file)
    # output is a list of PropertiesDict object with the following values:
    # [{'Title': 'Jumanji', 'Year': 1995},
    #  {'Title': 'Toy Story'}]
    ```

    * If `mode='all'` all fields for the content will be retrieved from raw source including the ones with a blank value

    ```title='JSON raw source'
    [{'Title': 'Jumanji', 'Year': 1995},
    {'Title': 'Toy Story', 'Year': ''}]
    ```

    ```python
    json_file = JSONFile(json_path)
    PropertiesFromDataset(mode='only_retrieved_evaluated').get_properties(json_file)
    # output is a list of PropertiesDict object with the following values:
    # [{'Title': 'Jumanji', 'Year': 1995},
    #  {'Title': 'Toy Story', 'Year': ''}]
    ```

    You could also choose exactly **which** fields to use to expand each content with the `field_name_list` parameter

    ```title='JSON raw source'
    [{'Title': 'Jumanji', 'Year': 1995},
    {'Title': 'Toy Story', 'Year': ''}]
    ```

    ```python
    json_file = JSONFile(json_path)
    PropertiesFromDataset(mode='only_retrieved_evaluated',
                          field_name_list=['Title']).get_properties(json_file)
    # output is a list of PropertiesDict object with the following values:
    # [{'Title': 'Jumanji'},
    #  {'Title': 'Toy Story'}]
    ```

    Args:

        mode: Parameter which specifies which properties should be retrieved.

            Possible values are ['only_retrieved_evaluated', 'all']:

                1. 'only retrieved evaluated' will retrieve properties which have a
                value, discarding ones with a blank value (i.e. '')
                2. 'all' will retrieve all properties, regardless if they have a value
                or not

        field_name_list: List of fields from the raw source that will be retrieved. Useful if you want to expand each
            content with only a subset of available properties from the local dataset

    """

    def __init__(self, mode: str = 'only_retrieved_evaluated', field_name_list: List[str] = None):
        super().__init__(mode)
        self.__field_name_list: List[str] = field_name_list

    def _check_mode(self, mode):
        modalities = {
            'all',
            'only_retrieved_evaluated',
        }
        if mode not in modalities:
            raise ValueError(f"mode={mode} not supported! Valid modalities are {modalities}")

    def get_properties(self, raw_source: RawInformationSource) -> List[PropertiesDict]:

        logger.info("Extracting exogenous properties from local dataset")
        prop_dict_list = []
        for raw_content in raw_source:

            if self.__field_name_list is None:
                prop_dict = raw_content
            else:
                prop_dict = {field: raw_content[field] for field in self.__field_name_list
                             if raw_content.get(field) is not None}

            if self.mode == 'only_retrieved_evaluated':
                prop_dict = {field: prop_dict[field] for field in prop_dict if prop_dict[field] != ''}

            prop_dict_list.append(PropertiesDict(prop_dict))

        return prop_dict_list

    def __str__(self):
        return "PropertiesFromDataset"

    def __repr__(self):
        return f'PropertiesFromDataset(mode={self.mode}, field_name_list={self.__field_name_list})'


class DBPediaMappingTechnique(ExogenousPropertiesRetrieval):
    """
    Exogenous technique which expands each content by using as external source the DBPedia ontology

    It needs the entity of the contents for which a mapping is required (e.g. entity_type=`dbo:Film`) and the field
    of the raw source that will be used for the actual mapping:


    Different modalities are available:

    * If `mode='only_retrieved_evaluated'`, all properties from DBPedia will be retrieved but discarding
    the ones with a blank value (i.e. '')

    * If `mode='all'`, all properties in DBPedia + all properties in local raw source will be retrieved.
    Local properties will be overwritten by dbpedia values if there's a conflict (same property in dbpedia and in local
    dataset)

    * If `mode='all_retrieved'`, all properties in DBPedia *only* will be retrieved

    * If `mode='original_retrieved'`, all local properties with their DBPedia value will be retrieved

    Args:
        entity_type: Domain of the contents you want to process (e.g. 'dbo:Film')
        label_field: Field of the raw source that will be used to map each content, DBPedia node with
            property **rdfs:label equal** to specified field value will be retrieved
        lang: Language of the `rdfs:label` that should match with `label_field` in the raw source
        mode: Parameter which specifies which properties should be retrieved.

            Possible values are ['only_retrieved_evaluated', 'all', 'all_retrieved', 'original_retrieved']:

                1. 'only retrieved evaluated' will retrieve properties which have a
                value, discarding ones with a blank value (i.e. '')
                2. 'all' will retrieve all properties from DBPedia + local source,
                regardless if they have a value or not
                3. 'all_retrieved' will retrieve all properties from DBPedia only
                4. 'original_retrieved' will retrieve all local properties with
                their DBPedia value
        return_prop_as_uri: If set to True, properties will be returned in their full uri form rather than in their
            rdfs:label form (e.g. "http://dbpedia.org/ontology/director" rather than "film director")
        max_timeout: Sometimes when mapping content to dbpedia, a batch of query may take longer than the max time
            allowed by the server due to internet issues: the framework will re-try the exact query `max_timeout` times
            before raising a `TimeoutError`
    """

    def __init__(self, entity_type: str, label_field: str, lang: str = 'EN',
                 mode: str = 'only_retrieved_evaluated', return_prop_as_uri: bool = False,
                 max_timeout: int = 5):
        super().__init__(mode)

        self._entity_type = entity_type
        self._label_field = label_field
        self._prop_as_uri = return_prop_as_uri
        self._lang = lang
        self._max_timeout = max_timeout

        self._sparql = SPARQLWrapper("https://dbpedia.org/sparql")
        self._sparql.setReturnFormat(JSON)

        self._class_properties = self._get_properties_class()

    def _query_dbpedia(self, query: str):

        timeout_counter = 0
        self._sparql.setQuery(query)
        self._sparql.setMethod(GET)

        while True:
            try:
                result = self._sparql.query()
                if result.response.status == 200:
                    timeout_counter = 0
                    break
                elif result.response.status == 206:
                    if timeout_counter >= self._max_timeout:
                        raise TimeoutError("Maximum number of trials reached!")

                    logger.warning(f"Timeout occurred! - {timeout_counter + 1} out of {self._max_timeout} possible \n"
                                   f"The timeout counter will be reset as soon as the request goes through")
                    timeout_counter += 1
                    query += '\n'  # add space to avoid query plan cache so that we perform a new request
                    self._sparql.setQuery(query)
            except URITooLong:
                self._sparql.setMethod(POST)

        return result.convert()["results"]["bindings"]

    def _get_uris_all_contents(self, raw_source: RawInformationSource):
        prefixes = "PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> "
        prefixes += "PREFIX dbo: <http://dbpedia.org/ontology/> "
        prefixes += "PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> "
        prefixes += "PREFIX foaf: <http://xmlns.com/foaf/0.1/> "

        all_contents_labels_original = [str(raw_content[self._label_field]) for raw_content in raw_source]

        select_clause = 'SELECT DISTINCT ?uri ?title '
        where_clause = 'WHERE { ' \
                       '?uri rdf:type dbo:Film . ' \
                       '?uri rdfs:label ?title . ' \
                       'BIND(str(?title) as ?str_label) . ' \
                       f'filter langMatches(lang(?title), "{self._lang}") '

        contents_taken = []
        results = []
        while True:
            contents_missing = set(all_contents_labels_original).difference(set(contents_taken))
            values_incomplete = ', '.join(f'"{wrapped}"' for wrapped in contents_missing)
            filter_query = f'filter(?str_label in ({values_incomplete}))'
            query_incomplete = prefixes + select_clause + where_clause + filter_query + "}"

            result_incomplete = self._query_dbpedia(query_incomplete)

            # if result_incomplete is empty, then there's no other item to map and we exit the loop
            if not result_incomplete:
                break

            results.extend(result_incomplete)
            contents_taken.extend([row["title"]["value"] for row in result_incomplete])

            logger.info(f"Contents mapped so far: {len(results)} of {len(all_contents_labels_original)} "
                        f"(ideal, tipically less are mapped)")

        logger.info(f"Contents mapped in total: {len(results)}")

        if len(results) == 0:
            raise ValueError("No mapping found")

        # Reset the order of contents
        res = {row["title"]["value"]: row["uri"]["value"] for row in results}
        uri_lables_dict = {'uri': [res[content] if res.get(content) else np.nan
                                   for content in all_contents_labels_original],
                           'label': all_contents_labels_original}

        results_df = pd.DataFrame.from_dict(uri_lables_dict)

        return results_df

    def _get_properties_class(self):
        query = "PREFIX dbo: <http://dbpedia.org/ontology/> "
        query += "SELECT DISTINCT ?property ?property_label WHERE { "
        query += "{ "
        query += "?property rdfs:domain ?class. "
        query += "%s rdfs:subClassOf+ ?class. " % self._entity_type
        query += "} UNION {"
        query += "?property rdfs:domain %s " % self._entity_type
        query += "} "
        query += "?property rdfs:label ?property_label. "
        query += f"FILTER (langMatches(lang(?property_label), \"{self._lang}\"))." + "} "

        results = self._query_dbpedia(query)

        if len(results) == 0:
            raise ValueError("The Entity type doesn't exists in DBPedia!")

        uri_labels_tuples = [(row["property"]["value"], row["property_label"]["value"])
                             for row in results]

        properties_df = pd.DataFrame.from_records(
            uri_labels_tuples,
            columns=['uri', 'label']
        )

        return properties_df

    def _retrieve_properties_contents(self, uris: pd.DataFrame):

        logger.info("Extracting properties for mapped contents...")

        query = "PREFIX dbo: <http://dbpedia.org/ontology/> "
        query += "SELECT ?uri ?property ?o WHERE {{ SELECT DISTINCT ?uri ?property ?o WHERE {"

        query += "VALUES ?property { "
        query += " ".join([f"<{uri_property}>" for uri_property in self._class_properties['uri']])
        query += "} "

        query += "VALUES ?uri { "
        query += " ".join([f"<{uri_item}>" for uri_item in uris['uri']])
        query += "} "

        query += "OPTIONAL {?uri ?property ?o . } "
        query += "} ORDER BY ?uri ?property ?o }} "

        results = []

        while True:
            # limit set at maximum allowed by dbpedia which is 10.000 may seem redundant
            # but the sparql endpoint won't work without it for this particular query
            query_incomplete = query + f" OFFSET {len(results)} LIMIT 10000"

            result_incomplete = self._query_dbpedia(query_incomplete)

            if not result_incomplete:
                break

            results.extend(result_incomplete)

        result_dict = defaultdict(lambda: defaultdict(list))
        for row in results:

            uri_item = row["uri"]["value"]
            property = row["property"]["value"]
            value = row.get("o", {}).get("value")  # safe way of accessing nested value in a dict

            # wrap every property value in a list, in case a property have more than one value
            # eg. starring: [DiCaprio, Tom Hardy]
            result_dict[uri_item][property].append(value)

        # index_map = {v: i for i, v in enumerate(list(uris['uri']))}
        # result_dict = dict(sorted(result_dict.items(), key=lambda pair: index_map[pair[0]]))

        # if some properties have only one value, then remove the list that wraps them
        # e.g. director: [Inarritu] -> director: Inarritu
        for uri in result_dict:
            result_dict[uri].update({prop: result_dict[uri][prop][0] for prop in result_dict[uri]
                                     if isinstance(result_dict[uri][prop], list) and len(result_dict[uri][prop]) == 1})

        return result_dict

    def _get_only_retrieved_evaluated(self, uris: pd.DataFrame, all_properties_dbpedia: dict) -> List[PropertiesDict]:

        prop_dict_list = []

        for uri in uris['uri']:

            if pd.notna(uri):

                content_properties_dbpedia = all_properties_dbpedia[uri]

                # Get only retrieved properties that have a value
                if self._prop_as_uri:
                    content_properties_final = {k: content_properties_dbpedia[k] for k in content_properties_dbpedia
                                                if content_properties_dbpedia[k] is not None}
                else:
                    # This goes into the class properties table and gets the labels to the corresponding uri
                    content_properties_final = {
                        self._class_properties.query('uri == @k')['label'].values[0]: content_properties_dbpedia[k]
                        for k in content_properties_dbpedia
                        if content_properties_dbpedia[k] is not None}

                prop_content = PropertiesDict(content_properties_final)
            else:
                prop_content = PropertiesDict({})

            prop_dict_list.append(prop_content)

        return prop_dict_list

    def _get_all_properties_retrieved(self, uris, all_properties_dbpedia) -> List[PropertiesDict]:
        prop_dict_list = []

        for uri in uris['uri']:

            if pd.notna(uri):

                content_properties_dbpedia = all_properties_dbpedia[uri]

                # Get all retrieved properties, so we substitute those with None with ""
                content_properties_final = {}
                for prop_uri in content_properties_dbpedia:

                    value = ''
                    if content_properties_dbpedia.get(prop_uri) is not None:
                        value = content_properties_dbpedia[prop_uri]

                    if self._prop_as_uri:
                        key = prop_uri
                    else:
                        # This goes into the class properties table and gets the labels to the corresponding uri
                        key = self._class_properties.query('uri == @prop_uri')['label'].values[0]

                    content_properties_final[key] = value

                prop_content = PropertiesDict(content_properties_final)
            else:
                prop_content = PropertiesDict({})

            prop_dict_list.append(prop_content)

        return prop_dict_list

    def __get_original_retrieved(self, uris, all_properties_dbpedia,
                                 raw_source: RawInformationSource) -> List[PropertiesDict]:

        prop_dict_list = []

        for uri, raw_content in zip(uris['uri'], raw_source):

            if pd.notna(uri):

                content_properties_dbpedia = all_properties_dbpedia[uri]
                content_properties_source = raw_content

                # Get all properties from source, those that have value in dbpedia will have value,
                # those that don't have value in dbpedia will be ''
                content_properties_final = {}
                for k in content_properties_source:
                    if k in self._class_properties['uri'].tolist():
                        value = content_properties_dbpedia[k] or ''

                        if self._prop_as_uri:
                            key = k
                        else:
                            key = self._class_properties.query('uri == @k')['label'].values[0]
                    elif k in self._class_properties['label'].tolist():
                        uri = self._class_properties.query('label == @k')['uri'].values[0]

                        value = content_properties_dbpedia[uri] or ''

                        if self._prop_as_uri:
                            key = uri
                        else:
                            key = k
                    else:
                        value = ''
                        key = k

                    content_properties_final[key] = value

                prop_content = PropertiesDict(content_properties_final)
            else:
                prop_content = PropertiesDict({})

            prop_dict_list.append(prop_content)

        return prop_dict_list

    def _get_all_properties(self, uris, all_properties_dbpedia,
                             raw_source: RawInformationSource) -> List[PropertiesDict]:
        prop_dict_list = []

        for uri, raw_content in zip(uris['uri'], raw_source):

            if pd.notna(uri):

                content_properties_dbpedia = all_properties_dbpedia[uri]
                content_properties_source = raw_content

                # Get all properties from source + all properties from dbpedia
                # if there are some properties in source that are also in dbpedia
                # the dbpedia value will overwrite the local source value
                content_properties_final = {}
                for k in content_properties_source:
                    if k in self._class_properties['uri'].tolist():
                        value = content_properties_dbpedia.pop(k)

                        if self._prop_as_uri:
                            key = k
                        else:
                            key = self._class_properties.query('uri == @k')['label'].values[0]

                    elif k in self._class_properties['label'].tolist():
                        uri = self._class_properties.query('label == @k')['uri'].values[0]

                        value = content_properties_dbpedia.pop(uri)

                        if self._prop_as_uri:
                            key = uri
                        else:
                            key = k
                    else:
                        value = content_properties_source[k]
                        key = k

                    if value is None:
                        value = content_properties_source[k]

                    content_properties_final[key] = value

                for k in content_properties_dbpedia:
                    value = content_properties_dbpedia[k]

                    if value is None:
                        value = ''

                    if self._prop_as_uri:
                        key = k
                    else:
                        key = self._class_properties.query('uri == @k')['label'].values[0]

                    content_properties_final[key] = value

                prop_content = PropertiesDict(content_properties_final)
            else:
                prop_content = PropertiesDict({})

            prop_dict_list.append(prop_content)

        return prop_dict_list

    def get_properties(self, raw_source: RawInformationSource) -> List[PropertiesDict]:
        logger.info("Extracting exogenous properties from DBPedia")
        uris = self._get_uris_all_contents(raw_source)

        uris_wo_none = uris.dropna()
        all_properties = self._retrieve_properties_contents(uris_wo_none)

        prop_dict_list = []
        if self.mode == 'only_retrieved_evaluated':
            prop_dict_list = self._get_only_retrieved_evaluated(uris, all_properties)

        elif self.mode == 'all_retrieved':
            prop_dict_list = self._get_all_properties_retrieved(uris, all_properties)

        elif self.mode == 'original_retrieved':
            prop_dict_list = self.__get_original_retrieved(uris, all_properties, raw_source)

        elif self.mode == 'all':
            prop_dict_list = self._get_all_properties(uris, all_properties, raw_source)

        return prop_dict_list

    def __str__(self):
        return "DBPediaMappingTechnique"

    def __repr__(self):
        return f'DBPediaMappingTechnique(mode={self.mode}, entity type={self._entity_type}, ' \
               f'label_field={self._label_field}, prop_as_uri={self._prop_as_uri}, max_timeout={self._max_timeout})'


class EntityLinking(ExogenousPropertiesRetrieval):
    """
    Abstract class that generalizes implementations that use entity linking for producing the semantic description
    """

    @abstractmethod
    def get_properties(self, raw_source: RawInformationSource) -> List[EntitiesProp]:
        raise NotImplementedError


class BabelPyEntityLinking(EntityLinking):
    """
    Exogenous technique which expands each content by using as external source the the BabelFy library.

    Each content will be expanded with the following babelfy properties (if available):

    * 'babelSynsetID',
    * 'DBPediaURL',
    * 'BabelNetURL',
    * 'score',
    * 'coherenceScore',
    * 'globalScore',
    * 'source'

    Args:
        field_to_link: Field of the raw source which will be used to search for the content properties in BabelFy
        api_key: String obtained by registering to babelfy website. If None only few queries can be executed
        lang: Language of the properties to retrieve
    """

    def __init__(self, field_to_link: str, api_key: str = None, lang: str = "EN"):
        super().__init__("all_retrieved")  # fixed mode since it doesn't make sense for babelfy
        self.__field_to_link = field_to_link
        self.__api_key = api_key
        self.__lang = lang
        self.__babel_client = BabelfyClient(self.__api_key, {"lang": lang})

    def get_properties(self, raw_source: RawInformationSource) -> List[EntitiesProp]:
        properties_list = []
        logger.info("Performing Entity Linking with BabelFy")
        with get_progbar(list(raw_source)) as pbar:
            for raw_content in pbar:
                data_to_disambiguate = check_not_tokenized(raw_content[self.__field_to_link])

                self.__babel_client.babelfy(data_to_disambiguate)

                properties_content = {}
                try:
                    if self.__babel_client.merged_entities is not None:

                        for entity in self.__babel_client.merged_entities:
                            properties_entity = {'babelSynsetID': '', 'DBPediaURL': '', 'BabelNetURL': '', 'score': '',
                                                 'coherenceScore': '', 'globalScore': '', 'source': ''}

                            for key in properties_entity:
                                if entity.get(key) is not None:
                                    properties_entity[key] = entity[key]

                            properties_content[entity['text']] = properties_entity

                    properties_list.append(EntitiesProp(properties_content))
                except AttributeError:
                    raise AttributeError("BabelFy limit reached! Insert an api key or change it if you inserted one!")

        return properties_list

    def __str__(self):
        return "BabelPyEntityLinking"

    def __repr__(self):
        return f'BabelPyEntityLinking(field_to_link={self.__field_to_link}, api_key={self.__api_key}, ' \
               f'lang={self.__lang})'



import inspect
import unittest

from clayrs.utils.automatic_methods import autorepr


class TestAutomaticMethods(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:

        # init method with only positional attributes
        class OnlyPositional:
            def __init__(self, attribute1, attribute2):
                self._repr_string = autorepr(self, inspect.currentframe())

            def __repr__(self):
                return self._repr_string

        cls.only_pos_class = OnlyPositional('formal1', 'formal2')

        # init method with only args attributes
        class OnlyArgs:
            def __init__(self, *args):
                self._repr_string = autorepr(self, inspect.currentframe())

            def __repr__(self):
                return self._repr_string

        cls.only_args_class = OnlyArgs('only_args1', 'only_args2')

        # init method with only kwargs attributes
        class OnlyKwargs:
            def __init__(self, **kwargs):
                self._repr_string = autorepr(self, inspect.currentframe())

            def __repr__(self):
                return self._repr_string

        cls.only_kwargs_class = OnlyKwargs(kwargs1='only_kwargs1', kwargs2='only_kwargs2')

        # init method with all possible attributes
        class AllPossibleArgs:
            def __init__(self, attribute1, attribute2, *args, **kwargs):
                self._repr_string = autorepr(self, inspect.currentframe())

            def __repr__(self):
                return self._repr_string

        cls.all_possible_args_class = AllPossibleArgs('formal1', 'formal2', 'args1', 'args2', 'args3',
                                                      kwargs1='kwargs_val')

    def test_autorepr(self):

        expected = "OnlyPositional(attribute1='formal1', attribute2='formal2')"
        result = repr(self.only_pos_class)

        self.assertEqual(expected, result)

        expected = "OnlyArgs(*args='only_args1', *args='only_args2')"
        result = repr(self.only_args_class)

        self.assertEqual(expected, result)

        expected = "OnlyKwargs(*kwargs_kwargs1='only_kwargs1', *kwargs_kwargs2='only_kwargs2')"
        result = repr(self.only_kwargs_class)

        self.assertEqual(expected, result)

        expected = "AllPossibleArgs(attribute1='formal1', attribute2='formal2', *args='args1', " \
                   "*args='args2', *args='args3', *kwargs_kwargs1='kwargs_val')"
        result = repr(self.all_possible_args_class)

        self.assertEqual(expected, result)


if __name__ == '__main__':
    unittest.main()



from . import content_based_algorithm
from . import graph_based_algorithm
from . import graphs

from .content_based_algorithm import *
from .graph_based_algorithm import *
from .graphs import *
from .recsys import ContentBasedRS, GraphBasedRS
from .partitioning import KFoldPartitioning, HoldOutPartitioning, BootstrapPartitioning
from .methodology import TestRatingsMethodology, TestItemsMethodology, TrainingItemsMethodology, AllItemsMethodology
from .experiment import ContentBasedExperiment, GraphBasedExperiment



from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING

import pandas as pd
import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error

if TYPE_CHECKING:
    from clayrs.recsys.partitioning import Split

from clayrs.evaluation.metrics.metrics import Metric


class ErrorMetric(Metric):
    """
    Abstract class for error metrics.
    An Error Metric evaluates 'how wrong' the recommender system was in predicting a rating
    """

    def perform(self, split: Split) -> pd.DataFrame:
        pred = split.pred
        truth = split.truth

        split_result = {'user_id': [], str(self): []}

        for user in set(truth.user_id_column):
            user_predictions = pred.get_user_interactions(user)
            user_truth = truth.get_user_interactions(user)

            # this will contain as key item id and as value the score in the truth of the user
            items_scores_truth = {truth_interaction.item_id: truth_interaction.score for truth_interaction in user_truth}

            # [(pred_score, truth_score), (pred_score, truth_score), ...]
            zipped_score_list = [(pred_interaction.score, items_scores_truth.get(pred_interaction.item_id))
                                 for pred_interaction in user_predictions
                                 if items_scores_truth.get(pred_interaction.item_id) is not None]

            if len(zipped_score_list) != 0:
                pred_score_list = [zipped_tuple[0] for zipped_tuple in zipped_score_list]
                truth_score_list = [zipped_tuple[1] for zipped_tuple in zipped_score_list]
                result = self._calc_metric(pred_score_list, truth_score_list)
            else:
                result = np.nan

            split_result['user_id'].append(user)
            split_result[str(self)].append(result)

        split_result['user_id'].append('sys')
        split_result[str(self)].append(np.nanmean(split_result[str(self)]))

        return pd.DataFrame(split_result)

    @abstractmethod
    def _calc_metric(self, truth_scores: list, pred_scores: list):
        """
        Private method that must be implemented by every error metric specifying how to calculate the metric
        for a single user given a Series of ratings taken from the ground truth and a Series of ratings predicted by the
        recommender system.

        Both Series must be in relation 1 to 1 between each other, meaning that if row 1 of the 'truth_scores' parameter
        contains the rating of the 'iPhone' item, row 1 of the 'pred_scores' parameter must contain the predicted rating
        of the 'iPhone' item, so both Series must be ordered in the same manner and must be filtered in order to exclude
        items which are in the ground truth of the user but are not predicted, as well as items which are predicted
        but aren't present in the ground truth. Debug the unittest cases for examples

        Args:
            truth_scores (pd.Series): Pandas Series which contains rating of the user taken from its ground truth. Has a
                1 to 1 relationship with the 'pred_scores' Series
            pred_scores (pd.Series): Pandas Series which contains rating predicted by the recommender system. Has a
                1 to 1 relationship with the 'truth_scores' Series
        """
        raise NotImplementedError


class MSE(ErrorMetric):
    r"""
    The MSE (Mean Squared Error) metric is calculated as such for the **single user**:

    $$
    MSE_u = \sum_{i \in T_u} \frac{(r_{u,i} - \hat{r}_{u,i})^2}{|T_u|}
    $$

    Where:

    - $T_u$ is the *test set* of the user $u$
    - $r_{u, i}$ is the actual score give by user $u$ to item $i$
    - $\hat{r}_{u, i}$ is the predicted score give by user $u$ to item $i$

    And it is calculated as such for the **entire system**:

    $$
    MSE_{sys} = \sum_{u \in T} \frac{MSE_u}{|T|}
    $$
    Where:

    - $T$ is the *test set*
    - $MSE_u$ is the MSE calculated for user $u$

    There may be cases in which some items of the *test set* of the user could not be predicted (eg. A CBRS was chosen
    and items were not present locally)

    In those cases the $MSE_u$ formula becomes

    $$
    MSE_u = \sum_{i \in T_u} \frac{(r_{u,i} - \hat{r}_{u,i})^2}{|T_u| - unk}
    $$

    Where:

    - $unk$ (*unknown*) is the number of items of the *user test set* that could not be predicted

    If no items of the user test set has been predicted ($|T_u| - unk = 0$), then:

    $$
    MSE_u = NaN
    $$
    """

    def __str__(self):
        return "MSE"

    def __repr__(self):
        return "MSE()"

    def _calc_metric(self, truth_scores: list, pred_scores: list):
        return mean_squared_error(truth_scores, pred_scores)


class RMSE(ErrorMetric):
    r"""
    The RMSE (Root Mean Squared Error) metric is calculated as such for the **single user**:

    $$
    RMSE_u = \sqrt{\sum_{i \in T_u} \frac{(r_{u,i} - \hat{r}_{u,i})^2}{|T_u|}}
    $$

    Where:

    - $T_u$ is the *test set* of the user $u$
    - $r_{u, i}$ is the actual score give by user $u$ to item $i$
    - $\hat{r}_{u, i}$ is the predicted score give by user $u$ to item $i$

    And it is calculated as such for the **entire system**:

    $$
    RMSE_{sys} = \sum_{u \in T} \frac{RMSE_u}{|T|}
    $$

    Where:

    - $T$ is the *test set*
    - $RMSE_u$ is the RMSE calculated for user $u$

    There may be cases in which some items of the *test set* of the user could not be predicted (eg. A CBRS was chosen
    and items were not present locally, a methodology different than *TestRatings* was chosen).

    In those cases the $RMSE_u$ formula becomes

    $$
    RMSE_u = \sqrt{\sum_{i \in T_u} \frac{(r_{u,i} - \hat{r}_{u,i})^2}{|T_u| - unk}}
    $$

    Where:

    - $unk$ (*unknown*) is the number of items of the *user test set* that could not be predicted

    If no items of the user test set has been predicted ($|T_u| - unk = 0$), then:

    $$
    RMSE_u = NaN
    $$
    """

    def __str__(self):
        return "RMSE"

    def __repr__(self):
        return f"RMSE()"

    def _calc_metric(self, truth_scores: list, pred_scores: list):
        return mean_squared_error(truth_scores, pred_scores, squared=False)


class MAE(ErrorMetric):
    r"""
    The MAE (Mean Absolute Error) metric is calculated as such for the **single user**:

    $$
    MAE_u = \sum_{i \in T_u} \frac{|r_{u,i} - \hat{r}_{u,i}|}{|T_u|}
    $$

    Where:

    - $T_u$ is the *test set* of the user $u$
    - $r_{u, i}$ is the actual score give by user $u$ to item $i$
    - $\hat{r}_{u, i}$ is the predicted score give by user $u$ to item $i$

    And it is calculated as such for the **entire system**:

    $$
    MAE_{sys} = \sum_{u \in T} \frac{MAE_u}{|T|}
    $$

    Where:

    - $T$ is the *test set*
    - $MAE_u$ is the MAE calculated for user $u$

    There may be cases in which some items of the *test set* of the user could not be predicted (eg. A CBRS was chosen
    and items were not present locally, a methodology different than *TestRatings* was chosen).

    In those cases the $MAE_u$ formula becomes

    $$
    MAE_u = \sum_{i \in T_u} \frac{|r_{u,i} - \hat{r}_{u,i}|}{|T_u| - unk}
    $$

    Where:

    - $unk$ (*unknown*) is the number of items of the *user test set* that could not be predicted

    If no items of the user test set has been predicted ($|T_u| - unk = 0$), then:

    $$
    MAE_u = NaN
    $$
    """

    def __str__(self):
        return "MAE"

    def __repr__(self):
        return f"MAE()"

    def _calc_metric(self, truth_scores: list, pred_scores: list):
        return mean_absolute_error(truth_scores, pred_scores)



from typing import Callable

import numpy as np
from unittest import TestCase
from math import isclose


class TestEmbeddingSource(TestCase):

    # Will be used by several test involving embeddings
    def assertWordEmbeddingMatches(self, source, embedding: np.ndarray, word: str):
        # 'similar_by_vector()' returns a list with top n
        # words similar to the vector given. I'm interested only in the most similar
        # so n = 1
        # for example, top_1 will be in the following form ("title", 1.0)
        top_1 = source.model.similar_by_vector(embedding, 1)[0]

        # So I'm using indices to access the tuples values.
        # 'like' contains how similar is 'embedding_word' to the 'embedding' vector given
        embedding_word = top_1[0]
        like = top_1[1]

        # if the word associated with the embedding vector returned by the model doesn't match the word passed as
        # argument, AssertionError is raised
        if not embedding_word == word:
            raise AssertionError("Word %s is not %s" % (embedding_word, word))

        # Obviously due to approximation the conversion won't return the
        # exact word, but if the likelihood it's equal to 1 with a maximum error of 'abs_tol'
        # I'm assuming it's exactly that word
        if not isclose(like, 1, abs_tol=1e-6):
            raise AssertionError("Word %s and result word %s do not match" % (embedding_word, word))



from unittest import TestCase

import os

from clayrs.content_analyzer.information_processor.nltk import NLTK
from clayrs.content_analyzer.raw_information_source import JSONFile
from test.content_analyzer.embeddings.test_embedding_source import TestEmbeddingSource
from clayrs.content_analyzer.embeddings.embedding_learner.doc2vec import GensimDoc2Vec
from clayrs.content_analyzer.embeddings.embedding_learner.fasttext import GensimFastText
from clayrs.content_analyzer.embeddings.embedding_learner.latent_semantic_analysis import GensimLatentSemanticAnalysis
from clayrs.content_analyzer.embeddings.embedding_learner.random_indexing import GensimRandomIndexing
from clayrs.content_analyzer.embeddings.embedding_learner.word2vec import GensimWord2Vec
from test import dir_test_files


file_path = os.path.join(dir_test_files, 'movies_info_reduced.json')
doc2vec_file_path = os.path.join(dir_test_files, "test_embedding_models/doc2vec_model.kv")
lsa_file_path = os.path.join(dir_test_files, "test_embedding_models/lsa/lsa_model.model")
ri_file_path = os.path.join(dir_test_files, "test_embedding_models/ri_model.model")
word2vec_file_path = os.path.join(dir_test_files, "test_embedding_models/word2vec_model.kv")
fasttext_file_path = os.path.join(dir_test_files, "test_embedding_models/fasttext_model.kv")


class TestEmbeddingLearner(TestCase):
    def test_extract_corpus(self):
        preprocessor = NLTK(stopwords_removal=True, stemming=True)
        fields = ["Title", "Released"]
        expected = [['jumanji', '15', 'dec', '1995'],
                    ['grumpier', 'old', 'men', '22', 'dec', '1995'],
                    ['toy', 'stori', '22', 'nov', '1995'],
                    ['father', 'bride', 'part', 'ii', '08', 'dec', '1995'],
                    ['heat', '15', 'dec', '1995'],
                    ['tom', 'huck', '22', 'dec', '1995'],
                    ['wait', 'exhal', '22', 'dec', '1995'],
                    ['sabrina', '15', 'dec', '1995'],
                    ['dracula', ':', 'dead', 'love', '22', 'dec', '1995'],
                    ['nixon', '05', 'jan', '1996'],
                    ['american', 'presid', '17', 'nov', '1995'],
                    ['goldeney', '17', 'nov', '1995'],
                    ['balto', '22', 'dec', '1995'],
                    ['cutthroat', 'island', '22', 'dec', '1995'],
                    ['casino', '22', 'nov', '1995'],
                    ['sudden', 'death', '22', 'dec', '1995'],
                    ['sens', 'sensibl', '26', 'jan', '1996'],
                    ['four', 'room', '25', 'dec', '1995'],
                    ['money', 'train', '22', 'nov', '1995'],
                    ['ace', 'ventura', ':', 'natur', 'call', '10', 'nov', '1995']]

        src = JSONFile(file_path)
        learner = GensimLatentSemanticAnalysis("./test_extract_corpus")
        generated = learner.extract_corpus(src, fields, [preprocessor])

        self.assertEqual(generated, expected)


class TestWordEmbeddingSourceGensimLearner(TestEmbeddingSource):
    def test_doc2vec(self):
        # model created using d2c_test_data.json
        source = GensimDoc2Vec(doc2vec_file_path)
        vector_size = 20
        result = source.load(["machine", "learning", "random_word"])

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(result[1].any(), True)
        self.assertEqual(result[2].any(), False)

        self.assertWordEmbeddingMatches(source, result[0], "machine")
        self.assertWordEmbeddingMatches(source, result[1], "learning")

    def test_fasttext(self):
        source = GensimFastText(fasttext_file_path)
        vector_size = 100
        result = source.load(["first", "remote"])

        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(result[1].any(), True)

        self.assertWordEmbeddingMatches(source, result[0], "first")
        self.assertWordEmbeddingMatches(source, result[1], "remote")

    def test_lsa(self):
        source = GensimLatentSemanticAnalysis(lsa_file_path)
        vector_size = source.get_vector_size()
        result = source.load(["first", "remote", "random_word"])

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(result[1].any(), True)
        self.assertEqual(result[2].any(), False)

    def test_ri(self):
        source = GensimRandomIndexing(ri_file_path)
        vector_size = 300
        result = source.load(["first", "remote", "random_word"])

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(result[1].any(), True)
        self.assertEqual(result[2].any(), False)

    def test_word2vec(self):
        source = GensimWord2Vec(word2vec_file_path)
        vector_size = 100
        result = source.load(["first", "exile", "random_word"])

        self.assertEqual(len(result), 3)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(result[1].any(), True)
        self.assertEqual(result[2].any(), False)

        self.assertWordEmbeddingMatches(source, result[0], "first")
        self.assertWordEmbeddingMatches(source, result[1], "exile")




from unittest import TestCase
from clayrs.content_analyzer.utils.check_tokenization import check_not_tokenized, check_tokenized, tokenize_in_sentences


class Test(TestCase):
    def test_check_tokenized(self):
        str_ = 'abcd efg'
        list_ = ['abcd', 'efg']
        check_tokenized(str_)
        check_tokenized(list_)
        check_not_tokenized(str_)
        check_not_tokenized(list_)

    def test_tokenize_sentence(self):

        phrases = "Ciao, questa è una prova. Anche questa. And this is the third"
        result = tokenize_in_sentences(phrases)

        self.assertTrue(len(result) == 3)


from typing import Type, Set
import inspect


def get_all_implemented_subclasses(cls: Type) -> Set:
    """
    Method that retrieves all implemented subclasses of a given class
    (also considering subclasses of a subclass and so on)

    The method calls itself to find the subclasses of each subclass

    Args:
        cls (Type): class from which all implemented subclasses will be extracted

    Returns:
        set containing all of cls' implemented subclasses
    """
    return set([sub for sub in cls.__subclasses__() if not inspect.isabstract(sub)]).union(
        [sub for c in cls.__subclasses__() for sub in get_all_implemented_subclasses(c) if not inspect.isabstract(sub)])


def get_all_implemented_classes(cls: Type) -> Set:
    """
    Method that retrieves all implemented subclasses of a given class
    (also considering subclasses of a subclass and so on)

    The method calls itself to find the subclasses of each subclass

    If the class passed as argument is not abstract, it is added to the set's results

    Args:
        cls (Type): class from which all implemented subclasses will be extracted

    Returns:
        set containing all of cls' implemented subclasses and cls itself if it is not abstract
    """

    classes = get_all_implemented_subclasses(cls)

    if not inspect.isabstract(cls):
        classes.add(cls)

    return classes



import gc
from typing import List
from abc import ABC, abstractmethod
import numpy as np


class EmbeddingSource(ABC):
    """
    General class whose purpose is to store the loaded pre-trained embeddings model and extract specified data from it

    The embedding source works in the following way: data -> model -> embedding vector
    The source iterates over all the data and returns the embedding vectors (in a matrix)

    EmbeddingSource has two classes that inherit from it: EmbeddingLoader and EmbeddingLearner. The first is used
    for models downloadable from the internet or stored locally (in particular, models that cannot be trained),
    the second is used for models stored locally that can be trained. Because of this, there
    shouldn't be any need for any other classes

    model: embeddings model loaded from source

    Args:
        reference (str): where to find the model, could be the model name to download or the path where the model is
            located
    """

    def __init__(self, reference: str):
        self.__reference = reference
        self.__model = None

    # this will load/download the model if not already loaded when called
    @property
    def model(self):
        if self.__model is None:
            try:
                self.__model = self.load_model() if self.__reference is not None else None
            except FileNotFoundError:
                self.__model = None
        return self.__model

    @property
    def reference(self):
        return self.__reference

    @model.setter
    def model(self, model):
        self.__model = model

    def load(self, text: List[str]) -> np.ndarray:
        """
        Function that extracts from the embeddings model the vectors of the data contained in text. If the model can't
        return a vector for the data passed as argument, then a vector filled with 0 will be created.

        Args:
            text (list<str>): text from which the embedding vectors will be extracted. The text contents depend from the
                embedding source's granularity. For example, a word embedding source will contain a list of tokens.

        Returns:
            embedding_matrix (np.ndarray): numpy vector, where each row represents the vector for the granularity (if
                the source has sentence granularity, the vector will refer to the corresponding sentence).
                Assuming text is a list of length N (where N depends by the granularity of the technique, so it could
                be the number of words or sentences), embedding_matrix will be N-dimensional.
        """
        if len(text) > 0:
            embedding_list = []
            for data in text:
                data = data.lower()
                try:
                    embedding_list.append(self.get_embedding(data))
                except KeyError:
                    embedding_list.append(np.zeros(self.get_vector_size()))
            embedding_matrix = np.asarray(embedding_list)
        else:
            # If the text is empty (eg. "") then the embedding matrix is a matrix
            # with 1 row filled with zeros
            embedding_matrix = np.zeros(shape=(1, self.get_vector_size()))

        return embedding_matrix

    @abstractmethod
    def load_model(self):
        """
        Method used to load the model. Each technique should implement this to define how the model is loaded
        """
        raise NotImplementedError

    def unload_model(self):
        self.__model = None
        gc.collect()

    @abstractmethod
    def get_vector_size(self) -> int:
        """
        Method that defines the size of a single embedding vector
        """
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, data) -> np.ndarray:
        """
        Method to return the embedding vector of the data passed as argument
        """
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List, Union, Callable, Optional, TYPE_CHECKING

from scipy.sparse import csr_matrix

if TYPE_CHECKING:
    from clayrs.content_analyzer.content_representation.content import FieldRepresentation

from clayrs.content_analyzer.content_representation.content import FeaturesBagField, SimpleField
from clayrs.content_analyzer.information_processor.information_processor import InformationProcessor
from clayrs.content_analyzer.raw_information_source import RawInformationSource
from clayrs.content_analyzer.utils.check_tokenization import check_not_tokenized


class FieldContentProductionTechnique(ABC):
    """
    Generic abstract class used to define the techniques that can be applied to the content's fields in order to
    produce their complex semantic representations.

    The FieldContentProductionTechnique creates, for each given content's raw data, the field's representation for a
    specific field
    """

    @staticmethod
    def process_data(data: str, preprocessor_list: List[InformationProcessor]) -> Union[List[str], str]:
        """
        The data passed as argument is processed using the preprocessor list (also given as argument) and is
        then returned

        Args:
            data (str): data on which each preprocessor, in the preprocessor list, will be used
            preprocessor_list (List[InformationProcessor]): list of preprocessors to apply to the data

        Returns:
            processed_data (Union[List[str], str): it could be a list of tokens created from the original data
            or the data in string form
        """
        processed_data = data
        for preprocessor in preprocessor_list:
            processed_data = preprocessor.process(processed_data)

        return processed_data

    @abstractmethod
    def produce_content(self, field_name: str, preprocessor_list: List[InformationProcessor],
                        source: RawInformationSource) -> List[FieldRepresentation]:
        """
        Abstract method that defines the methodology used by a technique to produce a list of FieldRepresentation.
        This list of FieldRepresentation objects contains all the complex representations referring to a specific
        field for each content. So if there were 3 contents (in their original/raw form) passed to the method in total,
        there would be 3 elements in the list (one for each content) and each one of these elements would be the
        representation generated by the technique. The data contained in the field for each content is also
        pre-processed by a list of information processors

        EXAMPLE

            contents to process: content1, content2, content3.
            each content contains a 'Plot' and a 'Title' field.
            if the goal is to create the representations for the field 'Plot' the method produce_content will take

            produce_content('Plot', [], raw data of the contents)

            this method will produce a list in the following form:

            [FieldRepresentation for content1, FieldRepresentation for content2, FieldRepresentation for content3]
            where each FieldRepresentation refers to the 'Plot' field of the designated content

            in this example the processor list didn't contain any information processor. If a processor list containing
            information processors was defined, the method would be:

            produce_content('Plot', [NLTK()], raw data of the contents)

            the NLTK() processor will be applied to the data contained in the field 'Plot' for each content and, after
            this process, the field representation will be elaborated.

        Args:
             field_name (str): name of the contents' field on which the technique will be applied
             preprocessor_list (List[InformationProcessor]): list of information processors that will pre-process the
                data contained in the field for each content
            source (RawInformationSource): source where the raw data of the contents is stored

        Returns:
            representation_list(List[FieldRepresentation]): list containing the representations generated by the
                technique for each content on a specific field
        """
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class SingleContentTechnique(FieldContentProductionTechnique):
    """
    Technique specialized in the production of representations that don't need any external information in order
    to be processed. This type of technique only considers the raw data within the content's field to create
    the complex representation
    """

    def produce_content(self, field_name: str, preprocessor_list: List[InformationProcessor],
                        source: RawInformationSource) -> List[FieldRepresentation]:
        """
        This method creates a list of FieldRepresentation objects, where each object is associated to a content
        and a specific field from the ones in said content. In order to do so, a simple preprocessing operation
        is done on the original data of the field (for each content) followed by the creation of the complex
        representation using the processed data. The complex representations are stored in a list and returned.
        """
        representation_list: List[FieldRepresentation] = []

        # it iterates over all contents contained in the source in order to retrieve the raw data
        # the data contained in the field_name is processed using each information processor in the processor_list
        # the data is passed to the method that will create the single representation
        for content_data in source:
            processed_data = self.process_data(content_data[field_name], preprocessor_list)
            representation_list.append(self.produce_single_repr(processed_data))

        return representation_list

    @abstractmethod
    def produce_single_repr(self, field_data: Union[List[str], str]) -> FieldRepresentation:
        """
        This method creates a single FieldRepresentation using the field data only (in the complete process made by the
        produce_content method, the data also goes through pre-processing)

        Args:
             field_data (str): data contained in a specific field

        Returns:
            FieldRepresentation: complex representation created using the field data
        """
        raise NotImplementedError


class CollectionBasedTechnique(FieldContentProductionTechnique):
    """
    Technique specialized in the production of representations that are in need of the entire collection in order
    to be processed. This type of technique performs a refactoring operation on the original dataset,
    so that each content in the collection is modified accordingly to the technique's needs
    """

    def produce_content(self, field_name: str, preprocessor_list: List[InformationProcessor],
                        source: RawInformationSource) -> List[FieldRepresentation]:
        """
        This method creates a list of FieldRepresentation objects, where each object is associated to a content
        and a specific field from the ones in said content. In order to do so, the entire dataset where the original
        contents are stored has to be modified (and also processed using the preproccesors). After that, it re-creates
        the content_id for each content and uses it to retrieve the content from the refactored dataset. Finally, the
        refactored dataset is deleted.
        """
        # the contents in the collection are modified by the technique
        # in this phase the data is also processed using the preprocessor_list
        dataset_len = self.dataset_refactor(source, field_name, preprocessor_list)

        representation_list: List[FieldRepresentation] = []

        # produces the representation, retrieving it from the dataset given the content's position in the refactored
        # dataset

        for i in range(0, dataset_len):
            representation_list.append(self.produce_single_repr(i))

        # once the operation is complete the refactored collection is deleted
        self.delete_refactored()

        return representation_list

    @abstractmethod
    def produce_single_repr(self, content_position: int) -> FieldRepresentation:
        """
        This method creates a single FieldRepresentation using the content position only. The content position is used
        to retrieve the corresponding content from the dataset modified during the content creation process.

        Args:
             content_position (int): position of the content in the dataset

        Returns:
            FieldRepresentation: complex representation created or retrieved from the refactored dataset using the
                content_position
        """
        raise NotImplementedError

    @abstractmethod
    def dataset_refactor(self, information_source: RawInformationSource, field_name: str,
                         preprocessor_list: List[InformationProcessor]) -> int:
        """
        This method restructures the raw data in a way functional to the final representation

        Args:
            information_source (RawInformationSource): source containing the raw data of the contents
            field_name (str): field interested in the content production process on which the refactor will be based on
            preprocessor_list (List[InformationProcessor]): list of preproccesors applied to the data in the field_name
                for each content during the refactoring operation

        Returns:
            length of the refactored dataset
        """
        raise NotImplementedError

    @abstractmethod
    def delete_refactored(self):
        """
        Once the content production phase is completed, the refactored dataset is useless. This method will delete the
        refactored dataset
        """
        raise NotImplementedError


class OriginalData(FieldContentProductionTechnique):
    """
    Technique used to retrieve the original data within the content's raw source without applying any
    processing operation.

    Note that if specified, preprocessing operations will still be applied!

    This technique is particularly useful if the user wants to keep the original
    data of the contents

    Args:
        dtype: If specified, data will be casted to the chosen dtype

    """

    def __init__(self, dtype: Callable = str):
        super().__init__()
        self.__dtype = dtype

    def produce_content(self, field_name: str, preprocessor_list: List[InformationProcessor],
                        source: RawInformationSource) -> List[SimpleField]:
        """
        The contents' raw data in the given field_name is extracted and stored in a SimpleField object.
        The SimpleField objects created are stored in a list which is then returned.
        No further operations are done on the data in order to keep it in the original form.
        Because of that the preprocessor_list is ignored and not used by this technique
        """

        representation_list: List[SimpleField] = []

        for content_data in source:
            processed_data = self.process_data(content_data[field_name], preprocessor_list)
            representation_list.append(SimpleField(self.__dtype(check_not_tokenized(processed_data))))

        return representation_list

    def __repr__(self):
        return f'OriginalData(dtype={self.__dtype})'

# DECODE POSSIBLE REPRESENTATION: Not implemented for now
#
# class DefaultTechnique(FieldContentProductionTechnique):
#     """
#     Default technique used when no FieldContentProductionTechnique is defined in the FieldConfig.
#     """
#
#     def __init__(self):
#         super().__init__()
#
#     def produce_content(self, field_name: str, preprocessor_list: List[InformationProcessor],
#                         source: RawInformationSource) -> List[FieldRepresentation]:
#         """
#         The content's raw data is decoded using the appropriate method (in case the data is not a string).
#         Each decoded representation is added to a list which is then returned
#         """
#         representation_list: List[FieldRepresentation] = []
#
#         for content_data in source:
#             # if a preprocessor is specified, then surely we must import the field data as a string,
#             # there's no other option
#             if len(preprocessor_list) != 0:
#                 representation = SimpleField(check_not_tokenized(self.process_data(str(content_data[field_name]),
#                                                                                    preprocessor_list)))
#
#             # If a preprocessor isn't specified, well maybe it is a complex representation:
#             # let's decode what kind of complex representation it is and import it accordingly.
#             else:
#                 representation = self.__decode_field_data(str(content_data[field_name]))
#
#             representation_list.append(representation)
#
#         return representation_list
#
#     def __decode_field_data(self, field_data: str):
#         # Decode string into dict or list
#         try:
#             loaded = json.loads(field_data)
#         except json.JSONDecodeError:
#             # in case the dict is {'foo': 1} json expects {"foo": 1}
#             reformatted_field_data = field_data.replace("\'", "\"")
#             try:
#                 loaded = json.loads(reformatted_field_data)
#             except json.JSONDecodeError:
#                 # if it has issues decoding we consider the data as str
#                 loaded = reformatted_field_data
#
#         # By default the representation decoded is what json tells us
#         decoded = SimpleField(loaded)
#
#         # if the decoded is a list, maybe it is an EmbeddingField repr
#         if isinstance(loaded, list):
#             arr = np.array(loaded)
#
#             # if the array values has can be converted into floats then we consider it as a dense vector
#             # else it is not and we do nothing
#             try:
#                 arr = arr.astype(float)
#                 decoded = EmbeddingField(arr)
#
#             # Can't be converted
#             except ValueError:
#                 pass
#
#         # if the decoded is a dict, maybe it is a FeaturesBagField
#         elif isinstance(loaded, dict):
#             # if all values of the dict are numbers or can be converted into numbers,
#             # then we consider it as a bag of words
#             # else it is not and we do nothing
#             if len(loaded.values()) != 0:
#                 try:
#                     dict_converted = {k: float(loaded[k]) for k in loaded}
#
#                     decoded = FeaturesBagField(dict_converted)
#
#                 # Can't be converted
#                 except ValueError:
#                     pass
#
#         return decoded


class TfIdfTechnique(CollectionBasedTechnique):
    """
    Abstract class that generalizes the implementations that produce a Bag of words with tf-idf metric
    """

    def __init__(self):
        self._tfidf_matrix: Optional[csr_matrix] = None
        self._feature_names: Optional[List[str]] = None

    def produce_single_repr(self, content_position: int) -> FeaturesBagField:
        """
        Retrieves the tf-idf values, for terms in document in the defined content_position,
        from the pre-computed word - document matrix.
        """
        nonzero_feature_index = self._tfidf_matrix[content_position, :].nonzero()[1]

        tfidf_sparse = self._tfidf_matrix.getrow(content_position).tocsc()
        pos_word_tuple = [(pos, self._feature_names[pos]) for pos in nonzero_feature_index]

        return FeaturesBagField(tfidf_sparse, pos_word_tuple)

    @abstractmethod
    def dataset_refactor(self, information_source: RawInformationSource, field_name: str,
                         preprocessor_list: List[InformationProcessor]):
        raise NotImplementedError

    def delete_refactored(self):
        del self._tfidf_matrix
        del self._feature_names


class SynsetDocumentFrequency(CollectionBasedTechnique):
    """
    Abstract class that generalizes implementations that use synsets
    """
    def __init__(self):
        self._synset_matrix: Optional[csr_matrix] = None
        self._synset_names: Optional[List[str]] = None

    def produce_single_repr(self, content_position: int) -> FeaturesBagField:
        """
        Retrieves the tf-idf values, for terms in document in the defined content_position,
        from the pre-computed word - document matrix.
        """
        nonzero_feature_index = self._synset_matrix[content_position, :].nonzero()[1]

        count_dense = self._synset_matrix.getrow(content_position).tocsc()
        pos_word_tuple = [(pos, self._synset_names[pos]) for pos in nonzero_feature_index]

        return FeaturesBagField(count_dense, pos_word_tuple)

    @abstractmethod
    def dataset_refactor(self, information_source: RawInformationSource, field_name: str,
                         preprocessor_list: List[InformationProcessor]):
        raise NotImplementedError

    def delete_refactored(self):
        del self._synset_matrix
        del self._synset_names

    def __repr__(self):
        return f'SynsetDocumentFrequency()'



import os
import unittest
from os import listdir
from os.path import splitext, isfile, join

from clayrs.content_analyzer import SearchIndex
from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict, LoadedContentsIndex
from test import dir_test_files


class TestLoadedContentsDict(unittest.TestCase):
    def test_all(self):
        # test load_available_contents for content based algorithm
        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')

        interface_dict = LoadedContentsDict(movies_dir)

        # we are testing get_contents_interface
        self.assertIsInstance(interface_dict.get_contents_interface(), dict)

        # since we didn't specify which items to load, we expected it has loaded all items from the folder
        expected = {splitext(filename)[0]
                    for filename in listdir(movies_dir)
                    if isfile(join(movies_dir, filename)) and splitext(filename)[1] == ".xz"}

        # we are testing also iter
        result = set(interface_dict)
        self.assertEqual(expected, result)

        # test loaded contents specified
        interface_dict = LoadedContentsDict(movies_dir, {'tt0112281', 'tt0112302'})

        # we are testing len
        self.assertTrue(len(interface_dict) == 2)

        # we are testing getitem
        self.assertIsNotNone(interface_dict['tt0112281'])
        self.assertIsNotNone(interface_dict['tt0112302'])

        # we are testing get
        self.assertIsNotNone(interface_dict.get('tt0112281'))
        self.assertIsNone(interface_dict.get('should be None'))


class TestLoadedContentsIndex(unittest.TestCase):
    def test_all(self):
        index = "../test/test_files/index"

        self.assertIsInstance(LoadedContentsIndex(index).get_contents_interface(), SearchIndex)



from unittest import TestCase
import os

from clayrs.content_analyzer import BertTransformers
from clayrs.content_analyzer.content_representation.content import EmbeddingField
from clayrs.content_analyzer.embeddings.embedding_learner import GensimFastText
from clayrs.content_analyzer.field_content_production_techniques.embedding_technique import \
    SentenceEmbeddingTechnique, Word2SentenceEmbedding, Word2DocEmbedding, \
    Sentence2DocEmbedding, WordEmbeddingTechnique, DocumentEmbeddingTechnique
from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique \
    import Centroid
from clayrs.content_analyzer.embeddings.embedding_loader.gensim import Gensim
from clayrs.content_analyzer.embeddings.embedding_loader.sbert import Sbert
from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.embedding_technique import \
    Sentence2WordEmbedding
from clayrs.content_analyzer.information_processor import NLTK
from clayrs.content_analyzer.raw_information_source import JSONFile
from test import dir_test_files

file_path = os.path.join(dir_test_files, "movies_info_reduced.json")


class TestEmbeddingTechnique(TestCase):
    def test_produce_content(self):
        technique = WordEmbeddingTechnique(GensimFastText())
        embedding_list = technique.produce_content("Plot", [NLTK()], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)

    def test_produce_content_str(self):
        self.skipTest("Test requires internet but is too complex to be mocked")
        technique = WordEmbeddingTechnique('glove-twitter-25')
        self.assertIsInstance(technique.embedding_source, Gensim)
        embedding_list = technique.produce_content("Plot", [NLTK()], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)

        technique = SentenceEmbeddingTechnique('paraphrase-distilroberta-base-v1')
        self.assertIsInstance(technique.embedding_source, Sbert)
        embedding_list = technique.produce_content("Plot", [NLTK()], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)

        technique = Word2DocEmbedding('glove-twitter-25', Centroid())
        self.assertIsInstance(technique.embedding_source, Gensim)
        embedding_list = technique.produce_content("Plot", [], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)

        technique = Sentence2DocEmbedding('paraphrase-distilroberta-base-v1', Centroid())
        self.assertIsInstance(technique.embedding_source, Sbert)
        embedding_list = technique.produce_content("Plot", [NLTK()], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)

        technique = Word2SentenceEmbedding('glove-twitter-25', Centroid())
        self.assertIsInstance(technique.embedding_source, Gensim)
        embedding_list = technique.produce_content("Plot", [], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)

    def test_load_not_existing_source(self):
        self.skipTest("Test requires internet but is too complex to be mocked")
        with self.assertRaises(FileNotFoundError):
            WordEmbeddingTechnique('not_existing_model')

        with self.assertRaises(FileNotFoundError):
            WordEmbeddingTechnique('./not_existing_path')

        with self.assertRaises(FileNotFoundError):
            SentenceEmbeddingTechnique('not_existing_model')

        with self.assertRaises(FileNotFoundError):
            SentenceEmbeddingTechnique('./not_existing_path')

        with self.assertRaises(FileNotFoundError):
            DocumentEmbeddingTechnique('not_existing_model')

        with self.assertRaises(FileNotFoundError):
            DocumentEmbeddingTechnique('./not_existing_path')

        with self.assertRaises(FileNotFoundError):
            WordEmbeddingTechnique(Gensim('not_existing_model'))


class TestSentenceEmbeddingTechnique(TestCase):

    def test_produce_single_repr(self):
        file_path = os.path.join(dir_test_files, "movies_info_reduced.json")
        fromsentencetowords = SentenceEmbeddingTechnique(BertTransformers("prajjwal1/bert-tiny"))

        embedding_list = fromsentencetowords.produce_content("Plot", [], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)


class TestFromSentenceWordsEmbeddingTechnique(TestCase):

    def test_produce_single_repr(self):
        file_path = os.path.join(dir_test_files, "movies_info_reduced.json")
        fromsentencetowords = Sentence2WordEmbedding(BertTransformers("prajjwal1/bert-tiny"))

        embedding_list = fromsentencetowords.produce_content("Plot", [], JSONFile(file_path))
        self.assertEqual(len(embedding_list), 20)
        self.assertIsInstance(embedding_list[0], EmbeddingField)



import os
from copy import deepcopy
from unittest import TestCase

import numpy as np
from scipy import sparse
from scipy.sparse import csr_matrix
from sklearn.feature_extraction import DictVectorizer

from clayrs.content_analyzer import Centroid
from clayrs.recsys import IndexQuery, LinearPredictor
from clayrs.recsys.content_based_algorithm.content_based_algorithm import ContentBasedAlgorithm
from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict, LoadedContentsIndex
from clayrs.utils.class_utils import get_all_implemented_subclasses
from clayrs.utils.load_content import load_content_instance

from clayrs.recsys.content_based_algorithm.centroid_vector.centroid_vector import CentroidVector
from clayrs.recsys.content_based_algorithm.centroid_vector.similarities import CosineSimilarity
from test import dir_test_files


class TestContentBasedAlgorithm(TestCase):

    def setUp(self) -> None:

        # ContentBasedAlgorithm is an abstract class, so we need to instantiate
        # a subclass to test its methods.
        self.alg = CentroidVector({'Plot': 'tfidf'}, CosineSimilarity(), 0)

    def test__bracket_representation(self):

        item_field = {'Plot': 'tfidf',
                      'Genre': [0],
                      'Title': [0, 'trybracket'],
                      'Director': 5}

        item_field_bracketed = {'Plot': ['tfidf'],
                                'Genre': [0],
                                'Title': [0, 'trybracket'],
                                'Director': [5]}

        result = self.alg._bracket_representation(item_field)

        self.assertEqual(item_field_bracketed, result)

    def test_extract_features_item(self):
        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')

        content = load_content_instance(movies_dir, 'tt0112281')

        result = self.alg.extract_features_item(content)

        self.assertEqual(1, len(result))
        self.assertIsInstance(result[0], sparse.csc_matrix)

    def test_fuse_representations(self):
        dv = DictVectorizer(sparse=False, sort=False)

        tfidf_result1 = {'word1': 1.546, 'word2': 1.467, 'word3': 0.55}
        doc_embedding_result1 = np.array([[0.98347, 1.384038, 7.1023803, 1.09854]])
        word_embedding_result1 = np.array([[0.123, 0.44561], [1.889, 3.22], [0.283, 0.887]])
        float_result1 = 8.8

        tfidf_result2 = {'word2': 1.467, 'word4': 1.1}
        doc_embedding_result2 = np.array([[2.331, 0.887, 1.1123, 0.7765]])
        word_embedding_result2 = np.array([[0.123, 0.44561], [5.554, 1.1234]])
        int_result2 = 7

        x = [[tfidf_result1, doc_embedding_result1, word_embedding_result1, float_result1],
             [tfidf_result2, doc_embedding_result2, word_embedding_result2, int_result2]]

        result = self.alg.fuse_representations(x, Centroid())

        dv.fit([tfidf_result1, tfidf_result2])
        centroid_word_embedding_1 = Centroid().combine(word_embedding_result1)
        centroid_word_embedding_2 = Centroid().combine(word_embedding_result2)

        expected_1 = np.hstack([dv.transform(tfidf_result1).flatten(), doc_embedding_result1.flatten(),
                                centroid_word_embedding_1.flatten(), float_result1])

        expected_2 = np.hstack([dv.transform(tfidf_result2).flatten(), doc_embedding_result2.flatten(),
                                centroid_word_embedding_2.flatten(), int_result2])

        self.assertTrue(all(isinstance(rep, np.ndarray) for rep in result))
        self.assertTrue(np.allclose(result[0], expected_1))
        self.assertTrue(np.allclose(result[1], expected_2))

    def test__load_available_contents(self):
        # test load_available_contents for content based algorithm
        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')

        interface_dict = self.alg._load_available_contents(movies_dir)
        self.assertIsInstance(interface_dict, LoadedContentsDict)

        interface_dict = self.alg._load_available_contents(movies_dir, {'tt0112281', 'tt0112302'})
        self.assertTrue(len(interface_dict) == 2)
        loaded_items_id_list = list(interface_dict)
        self.assertIn('tt0112281', loaded_items_id_list)
        self.assertTrue('tt0112302', loaded_items_id_list)

        # test load_available_contents for index
        index_alg = IndexQuery({'Plot': 'tfidf'})
        index_dir = os.path.join(dir_test_files, 'complex_contents', 'index')
        interface_dict = index_alg._load_available_contents(index_dir)
        self.assertIsInstance(interface_dict, LoadedContentsIndex)



from __future__ import annotations
import networkx as nx
from abc import ABC, abstractmethod
from typing import List, Union, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.recsys.graphs.graph import FullDiGraph
    from clayrs.recsys.graphs.graph import UserNode, ItemNode

from clayrs.recsys.graphs.feature_selection.exceptions import FeatureSelectionException
from clayrs.recsys.graphs.graph import PropertyNode


class FeatureSelectionAlgorithm(ABC):
    """
    Feature Selection algorithm class. It contains all the methods to use Feature Selection on a Full Graph.
    The main method to do so is the perform method.

    All the Feature Selection algorithms will create a new graph based on the one passed by the user. This new graph
    will contain only the Item or the User nodes from the original graph but the Property nodes will be different.
    Instead of having the property instance (for example: http://dbpedia.org/resource/Phil_Collins) the Property nodes
    in the new graph will have the property names associated to the property instance (for example: 'starring').
    This is retrieved because the label of the edge connecting a User or Item node to a Property node will be the

    Property name. So, for example:

    ```
    Item Node someitem -- (label: starring) --> Property Node http://dbpedia.org/resource/Phil_Collins
    ```

    will be transformed to:
    ```
    Item Node someitem ----> Property Node starring
    ```

    This is done so that only the instances related to the most important properties will be kept in the original graph

    After creating the new graph, an algorithm which can compute the 'importance' scores for each node in the new graph
    is used (for example, PageRank) and the Property nodes (in this case being the property labels of the original
    graph) with the best scores will be returned
    """

    def perform(self, graph: FullDiGraph, target_nodes: Union[List[UserNode], List[ItemNode]],
                mode: str = 'to_keep') -> list:
        """
        Main method that handles the entire Feature Selection process. After every step of the process is completed, it
        will return a list containing the most important property labels of the original graph if `mode='to_keep'` or
        it will return a list containing the least important property labels of the original graph if `mode='to_remove'`

        Args:
            graph: The original graph to apply Feature Selection on
            target_nodes: list of nodes from the original graph to consider for the creation
                of the new graph
            mode: String which governs what the method will return. With `mode='to_keep'` the method will return the
                the most important property labels, with `mode='to_remove'` the method will return the
                the least important property labels

        Returns:
            List of properties to keep or to remove of the original graph
        """
        raise NotImplementedError

    @staticmethod
    def create_new_graph(graph: FullDiGraph, target_nodes: Union[List[UserNode], List[ItemNode]]) -> nx.DiGraph:
        """
        Creates a NetworkX directed graph from the original graph and the target nodes list passed as argument

        Given links of this kind in the original graph:

        ```
        Item Node someitem -- (label: starring) --> Property Node http://dbpedia.org/resource/Phil_Collins
        ```

        This method will create a new graph where links will be transformed to:

        ```
        Item Node someitem ----> Property Node starring
        ```

        Args:
            graph (FullGraph): the original graph to apply Feature Selection on
            target_nodes (list[object]): list of user or item nodes from the original graph to consider for the creation
                of the new graph

        Returns:
            A new graph that will be created from the original graph and the list of its nodes to consider
        """
        new_graph = nx.DiGraph()

        # for each node in the target list, if it has any property node in the original graph, the target node will
        # be added together with all the property labels it is connected with in the graph. The property labels will be
        # turned into Property nodes in the new graph. A link between the target node and each new Property node will
        # be instantiated
        links_to_add = []
        for node in target_nodes:
            for successor_node in graph.get_successors(node):
                if isinstance(successor_node, PropertyNode):
                    link_data = graph.get_link_data(node, successor_node)
                    if link_data.get('weight') is None:
                        link = (node, PropertyNode(link_data['label']))
                    else:
                        link = (node, PropertyNode(link_data['label']), {'weight': link_data['weight']})

                    links_to_add.append(link)

        new_graph.add_edges_from(links_to_add)

        return new_graph


class TopKFeatureSelection(FeatureSelectionAlgorithm):
    """
    Feature Selection class that uses a top k strategy (meaning that the k properties with the highest score in the
    ranking will be considered as the most important properties)

    Args:
        k: Number of properties to keep
    """

    def __init__(self, k: int):
        if k < 0:
            raise ValueError("The number of properties to keep passed (k parameter) can't be negative!")
        self._k = k

    @abstractmethod
    def create_rank(self, graph: nx.DiGraph) -> dict:
        """
        Method that calculates the rank used for feature selection. The method will return a dictionary where the keys
        will be the nodes in the graph passed as argument and the values will be the 'importance' score for each
        node. This rank will be used by the Feature Selection process to select which nodes to keep and which to discard

        Args:
            graph (DiGraph): graph from which the rank for the nodes will be calculated

        Returns:
            Dictionary containing the 'importance' score for each node in the graph
        """
        raise NotImplementedError

    def perform(self, graph: FullDiGraph, target_nodes: Union[List[UserNode], List[ItemNode]],
                mode: str = 'to_keep') -> list:

        if mode != 'to_keep' and mode != 'to_remove':
            raise TypeError("The only valid modes are 'to_keep' and 'to_remove'!")

        new_graph = self.create_new_graph(graph, target_nodes)

        # if the algorithm is not able to converge, a FeatureSelectionException is thrown
        try:
            rank = self.create_rank(new_graph)
        except nx.NetworkXException as e:
            raise FeatureSelectionException(str(e))

        # only property nodes of the new graph are kept in the ranking
        rank = {node.value: rank[node] for node in rank if isinstance(node, PropertyNode)}

        # the ranking produced by the Feature Selection algorithm is sorted
        rank = dict(sorted(rank.items(), key=lambda item: item[1], reverse=True))

        if mode == 'to_keep':
            # all the nodes up to the k-th node in the rank will be returned
            # in a list. These need to be kept
            nodes_list = list(rank.keys())[:self._k]
        else:
            # all the nodes from the k-th node in the rank onwards will be returned
            # in a list. These need to be removed
            nodes_list = list(rank.keys())[self._k:]

        return nodes_list


class TopKPageRank(TopKFeatureSelection):
    """
    Computes the PageRank as FeatureSelection algorithm. Property labels of the original graph will be scored with their
    page rank score and only the top-k labels will be kept in the *feature selected graph*,
    while discarding the others

    Args:
        k: Top-k property labels to keep in the *feature selected graph*

        alpha: Damping parameter for PageRank, default=0.85.

        personalization: The "personalization vector" consisting of a dictionary with a key some subset of graph nodes
            and personalization value each of those. At least one personalization value must be non-zero.
            If not specfiied, a nodes personalization value will be zero.
            By default, a uniform distribution is used.

        max_iter: Maximum number of iterations in power method eigenvalue solver.

        tol: Error tolerance used to check convergence in power method solver.

        nstart: Starting value of PageRank iteration for each node.

        weight: Edge data key to use as weight.  If None weights are set to 1.

        dangling: The outedges to be assigned to any "dangling" nodes, i.e., nodes without any outedges.
            The dict key is the node the outedge points to and the dict value is the weight of that outedge.
            By default, dangling nodes are given outedges according to the personalization vector (uniform if not
            specified). This must be selected to result in an irreducible transition matrix
            (see notes under google_matrix). It may be common to have the dangling dict to be the same as the
            personalization dict.
    """

    def __init__(self, k: int = 10, alpha: Any = 0.85, personalization: Any = None, max_iter: Any = 100,
                 tol: Any = 1.0e-6, nstart: Any = None, weight: bool = True, dangling: Any = None):
        super().__init__(k)

        self.alpha = alpha
        self.personalization = personalization
        self.max_iter = max_iter
        self.tol = tol
        self.nstart = nstart
        self.weight = 'weight' if weight is True else None
        self.dangling = dangling

    def create_rank(self, graph: nx.DiGraph) -> dict:
        return nx.pagerank(graph, alpha=self.alpha, personalization=self.personalization,
                           max_iter=self.max_iter, tol=self.tol, nstart=self.nstart, weight=self.weight,
                           dangling=self.dangling)


class TopKEigenVectorCentrality(TopKFeatureSelection):
    """
    Computes the Eigen Vector Centrality as FeatureSelection algorithm. Property labels of the original graph will be
    scored with their eigen vector centrality score and only the top-k labels will be kept in the
    *feature selected graph*, while discarding the others

    Args:
        k: Top-k property labels to keep in the *feature selected graph*
        max_iter: Maximum number of iterations in power method.
        tol: Error tolerance used to check convergence in power method iteration.
        nstart: Starting value of eigenvector iteration for each node.
        weight: Boolean value which tells the algorithm if weight of the edges must be considered or not.
            Default is True
    """

    def __init__(self, k: int = 10, max_iter: Any = 100, tol: Any = 1.0e-6, nstart: Any = None, weight: bool = False):
        super().__init__(k)

        self.max_iter = max_iter
        self.tol = tol
        self.nstart = nstart
        self.weight = 'weight' if weight is True else None

    def create_rank(self, graph: nx.DiGraph) -> dict:
        return nx.eigenvector_centrality(graph, max_iter=self.max_iter, tol=self.tol,
                                         nstart=self.nstart, weight=self.weight)


class TopKDegreeCentrality(TopKFeatureSelection):
    """
    Computes the Degree Centrality as FeatureSelection algorithm. Property labels of the original graph will be
    scored with their degree centrality score and only the top-k labels will be kept in the *feature selected graph*,
    while discarding the others
    """

    def __init__(self, k: int = 10):
        super().__init__(k)

    def create_rank(self, graph: nx.DiGraph) -> dict:
        return nx.degree_centrality(graph)



from .linear_predictor import LinearPredictor
from .regressors import SkLinearRegression, SkRidge, SkBayesianRidge, SkSGDRegressor, SkARDRegression,\
    SkHuberRegressor, SkPassiveAggressiveRegressor



from __future__ import annotations
from typing import List, Tuple, Union, TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Prediction, Rank, Ratings
    from clayrs.evaluation.metrics.metrics import Metric

from clayrs.utils.context_managers import get_progbar


class MetricEvaluator:
    """
    Module of the Evaluation pipeline which, has the task to evaluate recommendations generated for every user with a
    list of metric specified

    This module can also be used to evaluate recommendations generated from outside. In this case, the usage is the
    following\n
    MANUAL USAGE::

        # Let's say we have outside recommendations for 2 splits, and
        # we want to evaluate them both

        # First we must wrap our generated recommendations into 'Split' objects
        split1 = Split(pred1, truth1)
        split2 = Split(pred2, truth2)

        # Then we instantiate the MetricCalculator passing the split list
        mc = MetricCalculator([split1, split2])

        # Then simply call the 'eval_metrics' method with the list of metrics you'd like to evaluate
        mc.eval_metrics([Precision(), Recall()])

    Args:
        predictions_truths (List[Split]): List of splits that need to be evaluated. Pass your custom splits here if you
            want to evaluate recommendations generated from the outside.
            This is set to None by default, meaning that splits that will be evaluated are those generated by
            the PredictionCalculator module of the EvalModel pipeline

    """

    # [ (total_pred, total_truth), (total_pred, total_truth) ...]
    def __init__(self, pred_list: Union[List[Prediction], List[Rank]], truth_list: List[Ratings]):

        self._pred_list = pred_list
        self._truth_list = truth_list

    def __repr__(self):
        return f'MetricEvaluator(pred_list={self._pred_list}, truth_list={self._truth_list})'

    def eval_metrics(self, metric_list: List[Metric]) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Method which effectively evaluates recommendations generated with the list of metric passed as argument.

        It returns two Pandas DataFrame, the first one containing system results on all metrics specified, the second
        one containing each users results for every metric eligible

        Args:
            metric_list (List[Metric]): List of metric on which recommendations need to be evaluated

        Returns:
            Two pandas DataFrame, the first will contain the system result for every metric specified inside the metric
            list, the second one will contain every user results for every metric eligible
        """

        frames_to_concat_users = []
        frames_to_concat_system = []

        with get_progbar(metric_list) as pbar:

            for metric in pbar:
                pbar.set_description(desc=f"Performing {metric}")

                metric_result_list = []

                for pred, truth in zip(self._pred_list, self._truth_list):
                    if len(pred) != 0 and len(truth) != 0:
                        user_id_valid = set(pred.user_id_column)
                        # Remove from truth users of which we do not have predictions
                        truth = truth.filter_ratings(user_id_valid)

                        metric_result = metric.perform(Split(pred, truth))

                        metric_result_list.append(metric_result)

                # if in future results for each fold for each user
                # set index as from_id and concat axis = 1
                total_results_metric = pd.concat(metric_result_list)

                if not total_results_metric.empty:
                    total_results_metric = total_results_metric.set_index('user_id')
                    system_results = total_results_metric.loc[['sys']]

                    # this means that if a system result is nan, it won't be present in the final df.
                    # Maybe there's a better solution?
                    # (This removal is done in case metric for users and system have different name eg. AP for users and
                    # MAP for sys)
                    system_results = system_results.dropna(axis='columns', how='all')

                    each_user_result = total_results_metric.drop(['sys'])
                    each_user_result = each_user_result.dropna(axis=1, how='all')

                    if not each_user_result.empty:
                        frames_to_concat_users.append(each_user_result)

                    if not system_results.empty:
                        frames_to_concat_system.append(system_results)

        # concat horizontally results of each metric both for users and system
        final_result_users = pd.DataFrame(columns=['user_id'])
        if len(frames_to_concat_users) != 0:
            final_result_users = pd.concat(frames_to_concat_users, axis=1)

            # for users calculate the mean
            final_result_users = final_result_users.groupby('user_id').mean()

        final_result_system = pd.DataFrame(columns=['user_id'])
        if len(frames_to_concat_system) != 0:
            final_result_system = pd.concat(frames_to_concat_system, axis=1)

            # replace index of system results in order to better identify results of each fold
            new_index = [f'sys - fold{i + 1}' for i in range(len(final_result_system))]
            final_result_system['user_id'] = new_index
            final_result_system = final_result_system.set_index('user_id')

            # add mean results as a row
            system_means = list(final_result_system.mean().values)
            final_result_system.loc['sys - mean'] = system_means

        return final_result_system, final_result_users


class Split:
    """
    Class container for two pandas DataFrame

    It may represent a split containing 'train set' and 'test set', or a split containing a ground truth and predictions
    for it, etc.

    Once instantiated, one can access the two dataframes in different ways:

    | > sp = Split()
    | > # Various ways of accessing the FIRST DataFrame
    | > sp.train
    | > sp.pred
    | > sp.first
    | >
    | > # Various ways of accessing the SECOND DataFrame
    | > sp.test
    | > sp.truth
    | > sp.second

    Args:
        first_set (pd.DatFrame): the first DataFrame to contain. If not specified, an empty DataFrame with 'from_id',
            'to_id', and 'score' column will be instantiated
        second_set (pd.DataFrame): the second DataFrame to contain. If not specified, an empty DataFrame with 'from_id',
            'to_id' and 'score' column will be instantiated
    """

    def __init__(self,
                 first_set: Ratings,
                 second_set: Ratings):

        self.__dict__['first'] = first_set
        self.__dict__['second'] = second_set

        self.__dict__['_valid_first_name'] = ['train', 'pred', 'first']
        self.__dict__['_valid_second_name'] = ['test', 'truth', 'second']

    def __getattr__(self, name):
        if name in self._valid_first_name:
            return self.first
        elif name in self._valid_second_name:
            return self.second

    def __setattr__(self, name, value):
        if name in self._valid_first_name:
            super().__setattr__('first', value)
        elif name in self._valid_second_name:
            super().__setattr__('second', value)



from unittest import TestCase

import os
import gensim

import numpy as np
from gensim.models import LsiModel

from clayrs.content_analyzer.embeddings.embedding_learner.latent_semantic_analysis import GensimLatentSemanticAnalysis
from gensim.corpora import Dictionary
from gensim.test.utils import common_texts

num_topics = 10
model_path = 'test_model_lda'


class TestLda(TestCase):
    def test_all(self):
        my_learner = GensimLatentSemanticAnalysis(model_path, num_topics=num_topics)

        corpus = common_texts
        my_learner.fit_model(corpus)

        # check that vector size is correct
        self.assertEqual(num_topics, my_learner.get_vector_size())

        common_dictionary = Dictionary(common_texts)
        common_corpus = [common_dictionary.doc2bow(text) for text in common_texts]
        expected_learner = LsiModel(common_corpus, num_topics=num_topics)

        # test get_embedding not existent document
        unseen_doc_text = ['this', 'is', 'a', 'new', 'document', 'which', 'doesnt', 'exist']

        # check that the doc is unseen (embedding has len 0)
        unseen_doc = common_dictionary.doc2bow(unseen_doc_text)
        expected = expected_learner[unseen_doc]
        self.assertTrue(len(expected) == 0)

        # in our framework if the doc is unseen KeyError is raised
        with self.assertRaises(KeyError):
            my_learner.get_embedding(unseen_doc_text)

        # test get_embedding existent document
        unseen_doc_text = ['human', 'time', 'trees']
        unseen_doc = common_dictionary.doc2bow(unseen_doc_text)
        expected = expected_learner[unseen_doc]
        expected_vector: np.ndarray = gensim.matutils.sparse2full(expected, num_topics)
        result_vector = my_learner.get_embedding(unseen_doc_text)

        # we don't have a way to check if the 2 vectors are the same, because they are build at random.
        # We just check that they are of the same length

        self.assertEqual(len(expected_vector), len(result_vector))

        # test save
        my_learner.save()
        self.assertTrue(os.path.isfile(f"{model_path}.model"))
        self.assertTrue(os.path.isfile(f"{model_path}.model.projection"))

        # test that after load we obtain a valid embedding
        my_learner_loaded = GensimLatentSemanticAnalysis(model_path)
        my_learner_loaded.load_model()
        unseen_doc_text = ['human', 'time', 'trees']
        result_vector = my_learner.get_embedding(unseen_doc_text)

        self.assertTrue(np.any(result_vector))




import os
import shutil
import unittest
from unittest import TestCase

import pandas as pd

from clayrs.content_analyzer import Ratings, Rank
from clayrs.evaluation.metrics.plot_metrics import LongTailDistr, PopRatioProfileVsRecs, PopRecsCorrelation
from clayrs.evaluation.eval_pipeline_modules.metric_evaluator import Split

original_ratings = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1',
                 'u2', 'u2', 'u2', 'u2', 'u2',
                 'u3', 'u3', 'u3', 'u3', 'u3', 'u3',
                 'u4', 'u4', 'u4', 'u4', 'u4', 'u4', 'u4',
                 'u5', 'u5', 'u5', 'u5', 'u5',
                 'u6', 'u6'],
     'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6', 'i7', 'i8',
                 'i1', 'i9', 'i10', 'i11', 'i55',
                 'i1', 'i12', 'i13', 'i3', 'i10', 'i14',
                 'i3', 'i10', 'i15', 'i16', 'i9', 'i17', 'i99',
                 'i10', 'i18', 'i19', 'i20', 'i21',
                 'inew_1', 'inew_2'],
     'score': [5, 4, 4, 1, 2, 3, 3, 1,
               4, 5, 1, 1, 1,
               3, 3, 2, 1, 1, 4,
               4, 4, 5, 5, 3, 3, 3,
               3, 3, 2, 2, 1,
               4, 3]})
original_ratings = Ratings.from_dataframe(original_ratings)

train = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1',  # removed last 2
                 'u2', 'u2', 'u2', 'u2',  # removed last 1
                 'u3', 'u3', 'u3', 'u3',  # removed last 2
                 'u4', 'u4', 'u4', 'u4', 'u4',  # removed last 2
                 'u5', 'u5', 'u5', 'u5',  # removed last 1
                 'u6'],  # removed last 1
     'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6',
                 'i1', 'i9', 'i10', 'i11',
                 'i1', 'i12', 'i13', 'i3',
                 'i3', 'i10', 'i15', 'i16', 'i9',
                 'i10', 'i18', 'i19', 'i20',
                 'inew_1'],
     'score': [5, 4, 4, 1, 2, 3,
               4, 5, 1, 1,
               3, 3, 2, 1,
               4, 4, 5, 5, 3,
               3, 3, 2, 2,
               4]})
train = Ratings.from_dataframe(train)

truth = pd.DataFrame(
    {'user_id': ['u1', 'u1',
                 'u2',
                 'u3', 'u3',
                 'u4', 'u4', 'u4',
                 'u5',
                 'u6'],
     'item_id': ['i7', 'i8',
                 'i55',
                 'i10', 'i14',
                 'i9', 'i17', 'i99',
                 'i21',
                 'inew_2'],
     'score': [3, 1,
               1,
               1, 4,
               3, 3, 3,
               1,
               3]})
truth = Ratings.from_dataframe(truth)

# u6 is missing, just to test DeltaGap in case for some users recs can't be computed
recs = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1',
                 'u2', 'u2', 'u2', 'u2', 'u2',
                 'u3', 'u3', 'u3', 'u3', 'u3',
                 'u4', 'u4', 'u4', 'u4', 'u4',
                 'u5', 'u5', 'u5', 'u5', 'u5'],
     'item_id': ['i7', 'i10', 'i11', 'i12', 'i13',
                 'i11', 'i20', 'i6', 'i3', 'i4',
                 'i4', 'i5', 'i6', 'i7', 'i10',
                 'i9', 'i2', 'i3', 'i1', 'i5',
                 'i2', 'i3', 'i4', 'i5', 'i6'],
     'score': [500, 400, 300, 200, 100,
               400, 300, 200, 100, 50,
               150, 125, 110, 100, 80,
               390, 380, 360, 320, 200,
               250, 150, 190, 100, 50]})
recs = Rank.from_dataframe(recs)

split = Split(recs, truth)


class TestLongTail(TestCase):
    def test_string_error(self):
        # Test invalid string passed
        with self.assertRaises(ValueError):
            LongTailDistr('.', on='invalid')

    def test_perform(self):
        # Save on same folder
        metric = LongTailDistr(on='truth')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_truth'
        self.assertTrue(os.path.isfile('./long_tail_distr_truth.png'))

        # Save on same folder with different format (svg)
        metric = LongTailDistr(on='truth', format='svg')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_truth'
        self.assertTrue(os.path.isfile('./long_tail_distr_truth.svg'))

        # Save on a not existent folder with a specified file_name
        metric = LongTailDistr('test_long', file_name='long_tail_custom_name', on='truth')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_truth'
        self.assertTrue(os.path.isfile(os.path.join('test_long', 'long_tail_custom_name_truth.png')))

        # Save on an existent folder with a specified file_name, long_tail based on pred frame
        metric = LongTailDistr('test_long', file_name='long_tail_custom_name', on='pred')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred'
        self.assertTrue(os.path.isfile(os.path.join('test_long', 'long_tail_custom_name_pred.png')))

        # Save on an existent folder with a filename already existent for the first time
        metric = LongTailDistr('test_long', file_name='long_tail_custom_name', on='pred')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred' + ' (1)' (Since it already exists)
        self.assertTrue(os.path.isfile(os.path.join('test_long', 'long_tail_custom_name_pred (1).png')))

        # Save on an existent folder with a filename already existent for the second time
        metric = LongTailDistr('test_long', file_name='long_tail_custom_name', on='pred')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred' + ' (2)' (Since it already exists)
        self.assertTrue(os.path.isfile(os.path.join('test_long', 'long_tail_custom_name_pred (2).png')))

    def test_overwrite(self):
        # Save on an existent folder with a specified file_name, long_tail based on pred frame
        metric = LongTailDistr('test_long/overwrite', file_name='long_tail_custom_name', on='pred')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred'
        self.assertTrue(os.path.isfile(os.path.join('test_long/overwrite', 'long_tail_custom_name_pred.png')))

        # Save on an existent folder with a filename already existent for the first time
        metric = LongTailDistr('test_long/overwrite', file_name='long_tail_custom_name', on='pred')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred' + ' (1)' (Since it already exists)
        self.assertTrue(os.path.isfile(os.path.join('test_long/overwrite', 'long_tail_custom_name_pred (1).png')))

        # Save on an existent folder with a filename already existent for the second time
        metric = LongTailDistr('test_long/overwrite', file_name='long_tail_custom_name', on='pred')
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred' + ' (2)' (Since it already exists)
        self.assertTrue(os.path.isfile(os.path.join('test_long/overwrite', 'long_tail_custom_name_pred (2).png')))

        # Save on an existent folder with a filename already existent and overwrite it
        metric = LongTailDistr('test_long/overwrite', file_name='long_tail_custom_name', on='pred', overwrite=True)
        metric.perform(split)
        # The graph is created with file_name = file_name + '_pred'
        self.assertTrue(os.path.isfile(os.path.join('test_long/overwrite', 'long_tail_custom_name_pred.png')))
        self.assertFalse(os.path.isfile(os.path.join('test_long/overwrite', 'long_tail_custom_name_pred (3).png')))

    @classmethod
    def tearDownClass(cls) -> None:
        os.remove('./long_tail_distr_truth.png')
        os.remove('./long_tail_distr_truth.svg')
        shutil.rmtree('test_long')


class TestPopRatioProfileVsRecs(TestCase):
    def test_perform(self):
        # Save on same folder
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings)
        metric.perform(split)
        self.assertTrue(os.path.isfile('./pop_ratio_profile_vs_recs.png'))

        # Save on same folder with a different format (svg)
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       format='svg')
        metric.perform(split)
        self.assertTrue(os.path.isfile('./pop_ratio_profile_vs_recs.svg'))

        # Save on a not existent folder with a specified file_name
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       out_dir='test_prof_recs',
                                       file_name='prof_vs_recs_custom_name')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs', 'prof_vs_recs_custom_name.png')))

        # Save on an existent folder with a specified file_name and a different pop_percentage
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       pop_percentage=0.6,
                                       out_dir='test_prof_recs',
                                       file_name='prof_vs_recs_different_percentage')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs', 'prof_vs_recs_different_percentage.png')))

        # Save on an existent folder with a specified file_name and with a different format (svg)
        # Save also the frame used to build the box_plot
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       out_dir='test_prof_recs',
                                       file_name='prof_vs_recs_svg_format',
                                       format='svg',
                                       store_frame=True)
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs', 'prof_vs_recs_svg_format.svg')))
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs', 'prof_vs_recs_svg_format.csv')))

    def test_perform_more_splits(self):
        # Save on same folder
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=[train, truth],
                                       original_ratings=original_ratings,
                                       out_dir='test_more_splits')
        metric.perform(split)
        self.assertTrue(os.path.isfile('test_more_splits/pop_ratio_profile_vs_recs.png'))

        # simulate 2 split call
        metric.perform(split)
        self.assertTrue(os.path.isfile('test_more_splits/pop_ratio_profile_vs_recs (1).png'))

        # simulate 3 split call, even if we only passed 2 user profiles frame
        # we will generate again the first plot
        metric.perform(split)
        self.assertTrue(os.path.isfile('test_more_splits/pop_ratio_profile_vs_recs (2).png'))
        self.assertEqual(os.path.getsize('test_more_splits/pop_ratio_profile_vs_recs.png'),
                         os.path.getsize('test_more_splits/pop_ratio_profile_vs_recs (2).png'))

    def test_overwrite(self):
        # Save on an existent folder with a specified file_name
        # Save also the frame used to build the box_plot
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       out_dir='test_prof_recs/overwrite',
                                       file_name='prof_vs_recs',
                                       store_frame=True)
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs.png')))
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs.csv')))

        # Save on an existent folder with an existent file_name for the first time
        # Save also the frame used to build the box_plot with an existent file_name for the first time
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       out_dir='test_prof_recs/overwrite',
                                       file_name='prof_vs_recs',
                                       store_frame=True)
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs (1).png')))
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs (1).csv')))

        # Save on an existent folder with an existent file_name for the second time
        # Save also the frame used to build the box_plot with an existent file_name for the second time
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       out_dir='test_prof_recs/overwrite',
                                       file_name='prof_vs_recs',
                                       store_frame=True)
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs (2).png')))
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs (2).csv')))

        # Save on an existent folder with an existent file_name and overwrite it
        # Save also the frame used to build the box_plot with an existent file_name and overwrite it
        metric = PopRatioProfileVsRecs(user_groups={'a': 0.5, 'b': 0.5},
                                       user_profiles=train,
                                       original_ratings=original_ratings,
                                       out_dir='test_prof_recs/overwrite',
                                       file_name='prof_vs_recs',
                                       store_frame=True,
                                       overwrite=True)
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs.png')))
        self.assertTrue(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs.csv')))

        self.assertFalse(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs (3).png')))
        self.assertFalse(os.path.isfile(os.path.join('test_prof_recs/overwrite', 'prof_vs_recs (3).csv')))

    # @classmethod
    # def tearDownClass(cls) -> None:
    #     os.remove('pop_ratio_profile_vs_recs.png')
    #     os.remove('pop_ratio_profile_vs_recs.svg')
    #     shutil.rmtree('test_prof_recs')
    #     shutil.rmtree('test_more_splits')


class TestPopRecsCorrelation(TestCase):
    def test_perform(self):
        # Save on same folder
        metric = PopRecsCorrelation(original_ratings=original_ratings)
        metric.perform(split)
        self.assertTrue(os.path.isfile('./pop_recs_correlation.png'))
        self.assertTrue(os.path.isfile('./pop_recs_correlation_no_zeros.png'))

        # Save on same folder with a different format (svg)
        metric = PopRecsCorrelation(original_ratings=original_ratings, format='svg')
        metric.perform(split)
        self.assertTrue(os.path.isfile('./pop_recs_correlation.svg'))
        self.assertTrue(os.path.isfile('./pop_recs_correlation_no_zeros.svg'))

        # Save on a not existent folder with a specified file_name
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs',
                                    file_name='pop_recs_custom_name')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs', 'pop_recs_custom_name.png')))
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs', 'pop_recs_custom_name_no_zeros.png')))

        # Save on an existent folder with a specified file_name and with a different format (svg)
        # Save also the frame used to build the box_plot
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs',
                                    file_name='pop_recs_custom_name', format='svg')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs', 'pop_recs_custom_name.svg')))
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs', 'pop_recs_custom_name_no_zeros.svg')))

    def test_perform_both(self):
        # Save both graph when it makes sense (eg. they are different)
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/both_yes',
                                    mode='both')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/both_yes', 'pop_recs_correlation.png')))
        # If The 'no-zeros' is created, its file_name will be: file_name = file_name + '_no_zeros'
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/both_yes', 'pop_recs_correlation_no_zeros.png')))

        truth = pd.DataFrame({
            'from_id': ['u1', 'u1', 'u2', 'u2', 'u2'],
            'to_id': ['i2', 'i1', 'i3', 'i5', 'i4'],
            'score': [5, 3, 3.6, 4, 2.2]}
        )
        truth = Ratings.from_dataframe(truth)

        recs = pd.DataFrame({
            'from_id': ['u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
            'to_id': ['i1', 'i2', 'inew1', 'inew2', 'i5', 'i4', 'i3'],
            'score': [300, 250, 200, 400, 350, 300, 100]}
        )
        recs = Ratings.from_dataframe(recs)
        # All items in the truth set have been recommended, so there's no 'zero' recommendation

        split_no_zero_present = Split(recs, truth)

        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/both_identical',
                                    mode='both')
        metric.perform(split_no_zero_present)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/both_identical', 'pop_recs_correlation.png')))
        # If The 'no-zeros' is created, its file_name will be: file_name = file_name + '_no_zeros'
        self.assertTrue(
            os.path.isfile(os.path.join('test_pop_recs/both_identical', 'pop_recs_correlation_no_zeros.png')))

    def test_perform_w_zeros(self):
        # Save only 'w-zeros' graph
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/w_zeros',
                                    mode='w_zeros')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/w_zeros', 'pop_recs_correlation.png')))
        # If The 'no-zeros' graph was created, its file_name would be: file_name = file_name + '_no_zeros'
        self.assertFalse(os.path.isfile(os.path.join('test_pop_recs/w_zeros', 'pop_recs_correlation_no_zeros.png')))

    def test_perform_no_zeros(self):
        # Save only 'no-zeros' graph
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/no_zeros',
                                    mode='no_zeros')
        metric.perform(split)
        self.assertFalse(os.path.isfile(os.path.join('test_pop_recs/no_zeros', 'pop_recs_correlation.png')))
        # The 'no-zeros' graph is created adding with file_name = file_name + '_no_zeros'
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/no_zeros', 'pop_recs_correlation_no_zeros.png')))

    def test_overwrite(self):
        # Save both graph when it makes sense (eg. they are different)
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/overwrite',
                                    mode='both')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation.png')))
        # If The 'no-zeros' is created, its file_name will be: file_name = file_name + '_no_zeros'
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation_no_zeros.png')))

        # Save both graph already existent for the first time
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/overwrite',
                                    mode='both')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation (1).png')))
        # If The 'no-zeros' is created, its file_name will be: file_name = file_name + '_no_zeros'
        self.assertTrue(
            os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation_no_zeros (1).png')))

        # Save both graph already existent for the second time
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/overwrite',
                                    mode='both')
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation (2).png')))
        # If The 'no-zeros' is created, its file_name will be: file_name = file_name + '_no_zeros'
        self.assertTrue(
            os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation_no_zeros (2).png')))

        # Save both graph already existent but overwrite them
        metric = PopRecsCorrelation(original_ratings=original_ratings,
                                    out_dir='test_pop_recs/overwrite',
                                    mode='both',
                                    overwrite=True)
        metric.perform(split)
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation.png')))
        # If The 'no-zeros' is created, its file_name will be: file_name = file_name + '_no_zeros'
        self.assertTrue(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation_no_zeros.png')))
        self.assertFalse(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation (3).png')))
        self.assertFalse(os.path.isfile(os.path.join('test_pop_recs/overwrite', 'pop_recs_correlation (3).png')))

    @classmethod
    def tearDownClass(cls) -> None:
        os.remove('./pop_recs_correlation.png')
        os.remove('./pop_recs_correlation.svg')
        os.remove('./pop_recs_correlation_no_zeros.png')
        os.remove('pop_recs_correlation_no_zeros.svg')
        shutil.rmtree('test_pop_recs')


if __name__ == '__main__':
    unittest.main()
